# importing_cnc25d.py
# it helps to let the test scripts find the cnc25d package
# created by charlyoleg on 2013/06/03
#
# (C) Copyright 2013 charlyoleg
#
# This file is part of the Cnc25D Python package.
# 
# Cnc25D is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Cnc25D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Cnc25D.  If not, see <http://www.gnu.org/licenses/>.


"""
importing_cnc25d completes sys.path to give access to the cnc25d package
"""

import sys, os
#getcwd_dir = os.getcwd()
#print("dbg191: getcwd_dir:", getcwd_dir)
file_dir= os.path.dirname(__file__)
#print("dbg192: file_dir:", file_dir)
#argv_dir= os.path.dirname(sys.argv[0])
#print("dbg193: argv_dir:", argv_dir)

test_script_dir=file_dir
if(test_script_dir==''):
  test_script_dir='.'
sys.path.append(test_script_dir+'/../..')
#sys.path.append('./../..') # this work only if you launch the script from its own directory




# gearwheel_macro.py
# the macro to generate a gearwheel.
# created by charlyoleg on 2013/06/19
#
# (C) Copyright 2013 charlyoleg
#
# This file is part of the Cnc25D Python package.
# 
# Cnc25D is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Cnc25D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Cnc25D.  If not, see <http://www.gnu.org/licenses/>.


################################################################
# this file intends being included in the file bin/cnc25d_example_generator.py
# for this purpose, there is some syntaxe restrictions
# don't use triple single-quotes (') and return character ('\'.'n') in this file
# but you can still use triple double-quote (")
################################################################

"""
this piece of code is an example of how to use the parametric design gearwheel
You can also use this file as a FreeCAD macro from the GUI
Look at the code. It's very simple to hack
"""

################################################################
# Installation pre-request
################################################################
# This script needs freecad and Cnc25D installed on your system
# visit those sites for more information:
# http://www.freecadweb.org/
# https://pypi.python.org/pypi/Cnc25D
#
# To install FreeCAD on Ubuntu, run the following command:
# > sudo apt-get install freecad
# or to get the newest version:
# > sudo add-apt-repository ppa:freecad-maintainers/freecad-stable
# > sudo apt-get update
# > sudo apt-get install freecad
# and optionally:
# >  sudo apt-get install freecad-doc freecad-dev
# To install the python package cnc25d, run the following command:
# > sudo pip install Cnc25D
# or
# > sudo pip install Cnc25D -U


################################################################
# header for Python / FreeCAD compatibility
################################################################

from cnc25d import cnc25d_api
cnc25d_api.importing_freecad()

#print("FreeCAD.Version:", FreeCAD.Version())

################################################################
# import
################################################################

#
import math
#
from cnc25d import gearwheel
#
import Part

    
################################################################
# parameters value
################################################################
#
# choose the values of the parameters by editing this file
# feature request : create a GUI with PyQt4 to edit those parameter values

##### from gear_profile
### first gear
# general
#gw_gear_type                      = gw_gear_type
gw_gear_tooth_nb                  = 18
gw_gear_module                    = 10.0
gw_gear_primitive_diameter        = 0
gw_gear_addendum_dedendum_parity  = 50.0
# tooth height
gw_gear_tooth_half_height           = 0 # 10.0
gw_gear_addendum_height_pourcentage = 100.0
gw_gear_dedendum_height_pourcentage = 100.0
gw_gear_hollow_height_pourcentage   = 25.0
gw_gear_router_bit_radius           = 3.0
# positive involute
gw_gear_base_diameter       = 0
gw_gear_force_angle         = 0
gw_gear_tooth_resolution    = 2
gw_gear_skin_thickness      = 0
# negative involute (if zero, negative involute = positive involute)
gw_gear_base_diameter_n     = 0
gw_gear_force_angle_n       = 0
gw_gear_tooth_resolution_n  = 0
gw_gear_skin_thickness_n    = 0
### second gear
# general
gw_second_gear_type                     = 'e'
gw_second_gear_tooth_nb                 = 25
gw_second_gear_primitive_diameter       = 0
gw_second_gear_addendum_dedendum_parity = 0 # 50.0
# tooth height
gw_second_gear_tooth_half_height            = 0
gw_second_gear_addendum_height_pourcentage  = 100.0
gw_second_gear_dedendum_height_pourcentage  = 100.0
gw_second_gear_hollow_height_pourcentage    = 25.0
gw_second_gear_router_bit_radius            = 0
# positive involute
gw_second_gear_base_diameter      = 0
gw_second_gear_tooth_resolution   = 0
gw_second_gear_skin_thickness     = 0
# negative involute (if zero, negative involute = positive involute)
gw_second_gear_base_diameter_n    = 0
gw_second_gear_tooth_resolution_n = 0
gw_second_gear_skin_thickness_n   = 0
### gearbar specific
gw_gearbar_slope                  = 0.0
gw_gearbar_slope_n                = 0.0
### position
# first gear position
gw_center_position_x                    = 0.0
gw_center_position_y                    = 0.0
gw_gear_initial_angle                   = 0.0
# second gear position
gw_second_gear_position_angle           = 0.0
gw_second_gear_additional_axis_length   = 0.0
### portion
#gw_portion_tooth_nb     = gw_portion_tooth_nb
#gw_portion_first_end    = gw_portion_first_end
#gw_portion_last_end     = gw_portion_last_end
### output
gw_gear_profile_height  = 20.0
gw_simulation_enable    = False    # gw_simulation_enable
#gw_output_file_basename = gw_output_file_basename
##### from gearwheel
### axle
gw_axle_type                = 'rectangle'
gw_axle_x_width             = 20.0
gw_axle_y_width             = 15.0
gw_axle_router_bit_radius   = 3.0
### wheel-hollow = legs
gw_wheel_hollow_leg_number        = 7
gw_wheel_hollow_leg_width         = 10.0
gw_wheel_hollow_leg_angle         = 0.0
gw_wheel_hollow_internal_diameter = 40.0
gw_wheel_hollow_external_diameter = 125.0
gw_wheel_hollow_router_bit_radius = 5.0
### cnc router_bit constraint
gw_cnc_router_bit_radius          = 1.0
### design output : view the gearwheel with tkinter or write files
gw_tkinter_view = True
gw_output_file_basename = "" # set a not-empty string if you want to generate the output files
#gw_output_file_basename = "test_output/gearwheel_macro.svg"  # to generate the SVG file with mozman svgwrite
#gw_output_file_basename = "test_output/gearwheel_macro.dxf"  # to generate the DXF file with mozman svgwrite
#gw_output_file_basename = "test_output/gearwheel_macro"      # to generate the Brep and DXF file with FreeCAD



################################################################
# action
################################################################

my_gw = gearwheel.gearwheel(
            ##### from gear_profile
            ### first gear
            # general
            #ai_gear_type                      = gw_gear_type,
            ai_gear_tooth_nb                  = gw_gear_tooth_nb,
            ai_gear_module                    = gw_gear_module,
            ai_gear_primitive_diameter        = gw_gear_primitive_diameter,
            ai_gear_addendum_dedendum_parity  = gw_gear_addendum_dedendum_parity,
            # tooth height
            ai_gear_tooth_half_height           = gw_gear_tooth_half_height,
            ai_gear_addendum_height_pourcentage = gw_gear_addendum_height_pourcentage,
            ai_gear_dedendum_height_pourcentage = gw_gear_dedendum_height_pourcentage,
            ai_gear_hollow_height_pourcentage   = gw_gear_hollow_height_pourcentage,
            ai_gear_router_bit_radius           = gw_gear_router_bit_radius,
            # positive involute
            ai_gear_base_diameter       = gw_gear_base_diameter,
            ai_gear_force_angle         = gw_gear_force_angle,
            ai_gear_tooth_resolution    = gw_gear_tooth_resolution,
            ai_gear_skin_thickness      = gw_gear_skin_thickness,
            # negative involute (if zero, negative involute = positive involute)
            ai_gear_base_diameter_n     = gw_gear_base_diameter_n,
            ai_gear_force_angle_n       = gw_gear_force_angle_n,
            ai_gear_tooth_resolution_n  = gw_gear_tooth_resolution_n,
            ai_gear_skin_thickness_n    = gw_gear_skin_thickness_n,
            ### second gear
            # general
            ai_second_gear_type                     = gw_second_gear_type,
            ai_second_gear_tooth_nb                 = gw_second_gear_tooth_nb,
            ai_second_gear_primitive_diameter       = gw_second_gear_primitive_diameter,
            ai_second_gear_addendum_dedendum_parity = gw_second_gear_addendum_dedendum_parity,
            # tooth height
            ai_second_gear_tooth_half_height            = gw_second_gear_tooth_half_height,
            ai_second_gear_addendum_height_pourcentage  = gw_second_gear_addendum_height_pourcentage,
            ai_second_gear_dedendum_height_pourcentage  = gw_second_gear_dedendum_height_pourcentage,
            ai_second_gear_hollow_height_pourcentage    = gw_second_gear_hollow_height_pourcentage,
            ai_second_gear_router_bit_radius            = gw_second_gear_router_bit_radius,
            # positive involute
            ai_second_gear_base_diameter      = gw_second_gear_base_diameter,
            ai_second_gear_tooth_resolution   = gw_second_gear_tooth_resolution,
            ai_second_gear_skin_thickness     = gw_second_gear_skin_thickness,
            # negative involute (if zero, negative involute = positive involute)
            ai_second_gear_base_diameter_n    = gw_second_gear_base_diameter_n,
            ai_second_gear_tooth_resolution_n = gw_second_gear_tooth_resolution_n,
            ai_second_gear_skin_thickness_n   = gw_second_gear_skin_thickness_n,
            ### gearbar specific
            ai_gearbar_slope                  = gw_gearbar_slope,
            ai_gearbar_slope_n                = gw_gearbar_slope_n,
            ### position
            # first gear position
            ai_center_position_x                    = gw_center_position_x,
            ai_center_position_y                    = gw_center_position_y,
            ai_gear_initial_angle                   = gw_gear_initial_angle,
            # second gear position
            ai_second_gear_position_angle           = gw_second_gear_position_angle,
            ai_second_gear_additional_axis_length   = gw_second_gear_additional_axis_length,
            ### portion
            #ai_portion_tooth_nb     = gw_cut_portion[0],
            #ai_portion_first_end    = gw_cut_portion[1],
            #ai_portion_last_end     = gw_cut_portion[2],
            ### output
            ai_gear_profile_height  = gw_gear_profile_height,
            ai_simulation_enable    = gw_simulation_enable,    # gw_simulation_enable,
            #ai_output_file_basename = gw_output_file_basename,
            ##### from gearwheel
            ### axle
            ai_axle_type                = gw_axle_type,
            ai_axle_x_width             = gw_axle_x_width,
            ai_axle_y_width             = gw_axle_y_width,
            ai_axle_router_bit_radius   = gw_axle_router_bit_radius,
            ### wheel-hollow = legs
            ai_wheel_hollow_leg_number        = gw_wheel_hollow_leg_number,
            ai_wheel_hollow_leg_width         = gw_wheel_hollow_leg_width,
            ai_wheel_hollow_leg_angle         = gw_wheel_hollow_leg_angle,
            ai_wheel_hollow_internal_diameter = gw_wheel_hollow_internal_diameter,
            ai_wheel_hollow_external_diameter = gw_wheel_hollow_external_diameter,
            ai_wheel_hollow_router_bit_radius = gw_wheel_hollow_router_bit_radius,
            ### cnc router_bit constraint
            ai_cnc_router_bit_radius          = gw_cnc_router_bit_radius,
            ### design output : view the gearwheel with tkinter or write files
            ai_tkinter_view = gw_tkinter_view,
            ai_output_file_basename = gw_output_file_basename)

#Part.show(my_gw)




# display_backend.py
# -*- coding: utf-8 -*-

# creates and displays the two Tkinter Canvas windows used by outline_backends.py
# created by charlyoleg on 2013/07/11
#
# (C) Copyright 2013 charlyoleg
#
# This file is part of the Cnc25D Python package.
# 
# Cnc25D is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Cnc25D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Cnc25D.  If not, see <http://www.gnu.org/licenses/>.


"""
display_backend.py creates and displays the two Tkinter Canvas windows used by outline_backends.py
"""


################################################################
# import
################################################################

import math
import sys, argparse
import Tkinter
#import tkMessageBox
import matplotlib.pyplot
import design_help # just for get_effective_args()

################################################################
# global variable
################################################################
initial_tkinter_canvas_width = 400
initial_tkinter_canvas_height = 400
tkinter_canvas_margin_x = 20
tkinter_canvas_margin_y = 20
g_fast_angle_speed = 1.0
g_step_angle_speed = float(g_fast_angle_speed)/8
g_slow_angle_speed = float(g_fast_angle_speed)/16
g_step_period = 100 # ms

################################################################
# ******** sub-functions for Two_Canvas ***********
################################################################

def find_outline_extremum(ai_outline_list):
  """ find the four extremum (min_x, min_y, max_x, max_y) of a list of outlines
  """
  # first point
  first_point_x = 0
  first_point_y = 0
  first_outline_type = ai_outline_list[0][0]
  first_outline = ai_outline_list[0][1]
  if((first_outline_type=='graphic_lines')or(first_outline_type=='overlay_lines')):
    first_point_x=first_outline[0][0]
    first_point_y=first_outline[0][1]
  elif((first_outline_type=='graphic_polygon')or(first_outline_type=='overlay_polygon')):
    first_point_x=first_outline[0]
    first_point_y=first_outline[1]
  else:
    print("ERR305: Error, the outline type is unknow: {:s}".format(first_outline_type))
    sys.exit(2)
  min_x=first_point_x
  max_x=first_point_x
  min_y=first_point_y
  max_y=first_point_y
  for outline in ai_outline_list:
    outline_type = outline[0]
    if((outline_type=='graphic_lines')or(outline_type=='overlay_lines')):
      lines = outline[1]
      for line in lines:
        #print("dbg772: line:", line)
        min_x = min(min_x, line[0], line[2])
        max_x = max(max_x, line[0], line[2])
        min_y = min(min_y, line[1], line[3])
        max_y = max(max_y, line[1], line[3])
    elif((outline_type=='graphic_polygon')or(outline_type=='overlay_polygon')):
      points = outline[1]
      for i in range(len(points)/2):
        #print("dbg773: i:", i)
        min_x = min(min_x, points[2*i])
        max_x = max(max_x, points[2*i])
        min_y = min(min_y, points[2*i+1])
        max_y = max(max_y, points[2*i+1])
  r_extremum = (min_x, min_y, max_x, max_y)
  #print("dbg653: r_extremum:", r_extremum)
  return(r_extremum)

def compute_scale_coef(ai_extremum, ai_canvas_size):
  """ compute the four scale coefficient to apply to the outlines to make it fit into a canvas
  """
  (min_x, min_y, max_x, max_y) = ai_extremum
  (canvas_width, canvas_height, margin_x, margin_y) = ai_canvas_size
  r_scale_coef = (1, 0, 1, 0)
  #print("dbg663:", max_x, min_x, max_y, min_y)
  if(((max_x-min_x)>0)and((max_y-min_y)>0)):
    lx = (canvas_width-2*margin_x)/float(max_x-min_x)
    ly = (canvas_height-2*margin_y)/float(max_y-min_y)
    lxy = min(lx, ly)
    # choose lx and ly to get (O,x,y) orthonormal direct
    lx = 1*lxy
    ly = -1*lxy
    #kx = margin_x - min_x*lx
    #ky = margin_y - min_y*ly
    kx = (canvas_width - (max_x-min_x)*lx)/2 - min_x*lx
    #ky = (canvas_height - (max_y-min_y)*ly)/2 - min_y*ly
    ky = (canvas_height - (max_y-min_y)*abs(ly))/2 - max_y*ly
    r_scale_coef = (lx, kx, ly, ky)
  return(r_scale_coef)

def scale_outline(ai_outline_list, ai_coef):
  """ apply the scale coefficient to a list of outlines
  """
  (lx, kx, ly, ky) = ai_coef
  r_outline_list = []
  for outline in ai_outline_list:
    outline_type = outline[0]
    if((outline_type=='graphic_lines')or(outline_type=='overlay_lines')):
      lines = outline[1]
      new_lines = []
      for line in lines:
        new_line = []
        for i in range(len(line)/2):
          new_line.append(line[2*i]*lx+kx)
          new_line.append(line[2*i+1]*ly+ky)
        new_lines.append(tuple(new_line))
      r_outline_list.append((outline_type, tuple(new_lines),  outline[2],  outline[3]))
    elif((outline_type=='graphic_polygon')or(outline_type=='overlay_polygon')):
      points = outline[1]
      new_points = []
      for i in range(len(points)/2):
        new_points.append(points[2*i]*lx+kx)
        new_points.append(points[2*i+1]*ly+ky)
      r_outline_list.append((outline_type, tuple(new_points),  outline[2],  outline[3], outline[4]))
  return(r_outline_list)

def compute_crop_limit(ai_selected_area, ai_canvas_size, ai_scale_coef):
  """ compute the crop four parameters according to the mouse positions, window size and current applied scale coefficient
  """
  (mouse_x1, mouse_y1, mouse_x2, mouse_y2) = ai_selected_area
  (canvas_width, canvas_height, margin_x, margin_y) = ai_canvas_size
  (lx, kx, ly, ky) = ai_scale_coef
  top_left_x = max(0, min(mouse_x1, mouse_x2))
  top_left_y = max(0, min(mouse_y1, mouse_y2))
  bottom_right_x = min(canvas_width, max(mouse_x1, mouse_x2))
  bottom_right_y = min(canvas_height, max(mouse_y1, mouse_y2))
  limit_x1=(top_left_x-kx)/float(lx)
  #limit_y1=(top_left_y-ky)/float(ly)
  limit_y2=(top_left_y-ky)/float(ly)
  limit_x2=(bottom_right_x-kx)/float(lx)
  #limit_y2=(bottom_right_y-ky)/float(ly)
  limit_y1=(bottom_right_y-ky)/float(ly)
  r_crop_limit = (limit_x1, limit_y1, limit_x2, limit_y2)
  return(r_crop_limit)

def crop_outline(ai_outline_list, ai_limit):
  """ crop a list of outlines to get a new list of outlines
  """
  (limit_x1, limit_y1, limit_x2, limit_y2) = ai_limit
  r_outline_list = []
  for outline in ai_outline_list:
    outline_type = outline[0]
    if((outline_type=='graphic_lines')or(outline_type=='overlay_lines')):
      lines = outline[1]
      new_lines = []
      for line in lines:
        include_new_line = 1
        for i in range(len(line)/2):
          if((line[2*i]<limit_x1)
              or (line[2*i]>limit_x2)
              or (line[2*i+1]<limit_y1)
              or (line[2*i+1]>limit_y2)):
            include_new_line = 0
        if(include_new_line==1):
          new_lines.append(line)
      if(len(new_lines)>0):
        r_outline_list.append((outline_type, tuple(new_lines),  outline[2],  outline[3]))
    elif((outline_type=='graphic_polygon')or(outline_type=='overlay_polygon')):
      points = outline[1]
      new_points = []
      for i in range(len(points)/2):
        if((points[2*i]>limit_x1)
            and (points[2*i]<limit_x2)
            and (points[2*i+1]>limit_y1)
            and (points[2*i+1]<limit_y2)):
          new_points.append(points[2*i])
          new_points.append(points[2*i+1])
        else:
          if(len(new_points)>1):
            r_outline_list.append((outline_type, tuple(new_points),  outline[2],  outline[3], outline[4]))
            new_points = []
      if(len(new_points)>1):
        r_outline_list.append((outline_type, tuple(new_points),  outline[2],  outline[3], outline[4]))
  return(r_outline_list)

################################################################
# ******** Two_Canvas class ***********
################################################################


class Two_Canvas():
  """ Tkinter frame work to display a gear system or other simple mechanism
      It contains:
      - a main window with a graphic overview and some button control
      - a zoom window
      - a simple input parameter window
      - a matplotlib window to display curves
  """

  def action_button_fast_backward(self):
    """ widget action
    """
    self.rotation_direction = -1
    self.angle_speed = self.rotation_direction*g_fast_angle_speed
    #self.angle_position += self.angle_speed
    self.set_label_content()

  def action_button_slow_backward(self):
    """ widget action
    """
    self.rotation_direction = -1
    self.angle_speed = self.rotation_direction*g_slow_angle_speed
    #self.angle_position += self.angle_speed
    self.set_label_content()

  def action_button_step_backward(self):
    """ widget action
    """
    self.rotation_direction = -1
    self.angle_speed = 0
    self.angle_position += self.rotation_direction*g_step_angle_speed
    self.set_label_content()

  def action_button_stop(self):
    """ widget action
    """
    #self.rotation_direction = self.rotation_direction
    self.angle_speed = 0
    #self.angle_position += 0
    self.set_label_content()

  def action_button_step_forward(self):
    """ widget action
    """
    self.rotation_direction = 1
    self.angle_speed = 0
    self.angle_position += self.rotation_direction*g_step_angle_speed
    self.set_label_content()

  def action_button_slow_forward(self):
    """ widget action
    """
    self.rotation_direction = 1
    self.angle_speed = self.rotation_direction*g_slow_angle_speed
    #self.angle_position += self.angle_speed
    self.set_label_content()

  def action_button_fast_forward(self):
    """ widget action
    """
    self.rotation_direction = 1
    self.angle_speed = self.rotation_direction*g_fast_angle_speed
    #self.angle_position += self.angle_speed
    self.set_label_content()

  def set_label_content(self):
    """ Update the angle labels
    """
    lb_rotation_direction = '+' if (self.rotation_direction==1) else '-'
    lb1 = "Angle position : {:0.3f} radian   {:0.2f} degree".format(self.angle_position, self.angle_position*180/math.pi)
    lb2 = "Angle speed    : {:0.3f} radian   {:0.2f} degree  {:s}".format(self.angle_speed, self.angle_speed*180/math.pi, lb_rotation_direction)
    #print("dbg774: lb1:", lb1)
    #print("dbg775: lb2:", lb2)
    self.label1_content.set(lb1)
    self.label2_content.set(lb2)
    #self.label3_content.set(lb3)
    #print("dbg845: self.label1_content.get:", self.label1_content.get())

  def action_button_check_event(self, event):
    """ This function is only used for debug
        It has no utility for the Two_Canvas class
    """
    print("dbg141: event:", event)
    print("dbg142: event.time:", event.time)
    print("dbg143: event.type:", event.type)
    print("dbg144: event.widget:", event.widget)
    print("dbg145: event.keysym:", event.keysym)

  def action_canvas_a_mouse_button_press(self, event):
    """ Start of the zoom area selection
    """
    self.mouse_x1 = event.x
    self.mouse_y1 = event.y
    self.mouse_x2 = self.mouse_x1
    self.mouse_y2 = self.mouse_y1
    self.canvas_a_mouse_press = 1
    #print("dbg874: mouse_x1, mouse_y1:", self.mouse_x1, self.mouse_y1)

  def action_canvas_a_mouse_button_motion(self, event):
    """ widget action
    """
    self.mouse_x2 = event.x
    self.mouse_y2 = event.y
    #self.canvas_a.create_rectangle(self.mouse_x1, self.mouse_y1, self.mouse_x2, self.mouse_y2, fill='', outline='red', width=2)

  def action_canvas_a_mouse_button_release(self, event):
    """ End of the zoom area selection
        It computes the crop_limit parameters
    """
    self.mouse_x2 = event.x
    self.mouse_y2 = event.y
    self.canvas_a_mouse_press = 0
    #print("dbg875: mouse_x2, mouse_y2:", self.mouse_x2, self.mouse_y2)
    selected_area = (self.mouse_x1, self.mouse_y1, self.mouse_x2, self.mouse_y2)
    canvas_a_size = (self.canvas_a.winfo_width(), self.canvas_a.winfo_height(), tkinter_canvas_margin_x, tkinter_canvas_margin_y)
    if(self.scale_coef_a==None):
      print("Warn763: Warning, self.scale_coef_a is not set!")
    else:
      self.crop_limit = compute_crop_limit(selected_area, canvas_a_size, self.scale_coef_a)
      #print("dbg986: self.crop_limit:", self.crop_limit)

  def action_canvas_b_mouse_button_press(self, event):
    """ Start of the length measurement in canvas_b
    """
    self.mouse_bx1 = event.x
    self.mouse_by1 = event.y
    self.mouse_bx2 = self.mouse_bx1
    self.mouse_by2 = self.mouse_by1
    self.canvas_b_mouse_press = 1
    #print("dbg877: mouse_bx1, mouse_by1:", self.mouse_bx1, self.mouse_by1)

  def action_canvas_b_mouse_button_motion(self, event):
    """ widget action in canvas_b
    """
    self.mouse_bx2 = event.x
    self.mouse_by2 = event.y

  def action_canvas_b_mouse_button_release(self, event):
    """ End of the length measurement
        It computes and display the length in the log
    """
    self.mouse_bx2 = event.x
    self.mouse_by2 = event.y
    self.canvas_b_mouse_press = 0
    #print("dbg878: mouse_bx2, mouse_by2:", self.mouse_bx2, self.mouse_by2)
    if(self.scale_coef_b==None):
      print("WARN663: Warning, scale_coef_b is not computed yet!")
    else:
      (lx, kx, ly, ky) = self.scale_coef_b
      p1x = (self.mouse_bx1-kx)/lx
      p1y = (self.mouse_by1-ky)/ly
      p2x = (self.mouse_bx2-kx)/lx
      p2y = (self.mouse_by2-ky)/ly
      length_measurement = math.sqrt((p2x-p1x)**2+(p2y-p1y)**2)
      if(length_measurement>0):
        angle_measurement = math.atan2((p2y-p1y)/length_measurement, (p2x-p1x)/length_measurement)
        self.measurement_id += 1
        print("Info: measurement {:2d} : lenght: {:0.3f}  angle: {:0.3f}".format(self.measurement_id, length_measurement, angle_measurement))
        print("P1x: {:0.2f}  P1y: {:0.2f}  P2x: {:0.2f}  P2y: {:0.2f}".format(p1x, p1y, p2x, p2y))

  def simulation_step(self):
    """ Time simulation main function
    """
    self.angle_position += self.angle_speed
    self.set_label_content()
    self.apply_canvas_graphic_function()
    #self.apply_curve_graphic_table(self.angle_position, self.angle_speed)
    self.frame_a.after(g_step_period, self.simulation_step)
  
  def draw_canvas(self, ai_canvas, ai_canvas_graphics, ai_overlay):
    """ Draw the computed outline list into a canvas
    """
    for outline in ai_canvas_graphics:
      outline_type = outline[0]
      if(outline_type=='graphic_lines'):
        lines = outline[1]
        for line in lines:
          ai_canvas.create_line(line, fill=outline[2], width=outline[3])
      elif(outline_type=='overlay_lines'):
        if(ai_overlay==1):
          lines = outline[1]
          for line in lines:
            ai_canvas.create_line(line, fill=outline[2], width=outline[3])
      elif(outline_type=='graphic_polygon'):
        ai_canvas.create_polygon(outline[1], fill=outline[2], outline=outline[3], width=outline[4])
      elif(outline_type=='overlay_polygon'):
        if(ai_overlay==1):
          ai_canvas.create_polygon(outline[1], fill=outline[2], outline=outline[3], width=outline[4])

  def apply_canvas_graphic_function(self):
    """ compute and draw the outline list for the main and the zoom window
    """
    canvas_a_width = self.canvas_a.winfo_width()
    canvas_a_height = self.canvas_a.winfo_height()
    canvas_b_width = self.canvas_b.winfo_width()
    canvas_b_height = self.canvas_b.winfo_height()
    #print("dbg104: canvas_a_width:", canvas_a_width)
    #print("dbg105: canvas_a_height:", canvas_a_height)
    #if(self.canvas_graphic_function==None):
    #  #print("ERR446: Error, the canvas_graphic_function has not been set!")
    #  #sys.exit(2)
    #  print("WARN446: Warning, the canvas_graphic_function has not been set!")
    if(self.canvas_graphic_function!=None):
      all_graphics = self.canvas_graphic_function(self.rotation_direction, self.angle_position)
      #
      self.canvas_a.delete(Tkinter.ALL)
      # uncomment if you want to scale outline depending on the angle_position
      #self.outline_extremum = find_outline_extremum(all_graphics)
      canvas_a_size = (canvas_a_width, canvas_a_height, tkinter_canvas_margin_x, tkinter_canvas_margin_y)
      self.scale_coef_a = compute_scale_coef(self.outline_extremum, canvas_a_size)
      canvas_a_graphics = scale_outline(all_graphics, self.scale_coef_a)
      self.draw_canvas(self.canvas_a, canvas_a_graphics, self.overlay)
      if(self.canvas_a_mouse_press==1):
        self.canvas_a.create_rectangle(self.mouse_x1, self.mouse_y1, self.mouse_x2, self.mouse_y2, fill='', outline='red', width=2)
      #
      if(self.crop_limit==(0,0,0,0)):
        self.crop_limit=(self.outline_extremum[0], self.outline_extremum[1], (self.outline_extremum[0]+self.outline_extremum[2])/2, (self.outline_extremum[1]+self.outline_extremum[3])/2)
      #
      self.canvas_b.delete(Tkinter.ALL)
      crop_graphics = crop_outline(all_graphics, self.crop_limit)
      #print("dbg857: len(crop_graphics):", len(crop_graphics))
      canvas_b_size = (canvas_b_width, canvas_b_height, 2, 2)
      #print("dbg768: self.crop_limit:", self.crop_limit)
      #print("dbg763: canvas_b_size:", canvas_b_size)
      self.scale_coef_b = compute_scale_coef(self.crop_limit, canvas_b_size)
      #print("dbg854: scale_coef_b:", scale_coef_b)
      canvas_b_graphics = scale_outline(crop_graphics, self.scale_coef_b)
      #print("dbg986: canvas_b_graphics:", canvas_b_graphics)
      self.draw_canvas(self.canvas_b, canvas_b_graphics, self.overlay)
      #self.canvas_b.create_line((5,5,canvas_b_width-5,canvas_b_height-5), fill='yellow', width=3)
      if(self.canvas_b_mouse_press==1):
        self.canvas_b.create_line(self.mouse_bx1, self.mouse_by1, self.mouse_bx2, self.mouse_by2, fill='red', width=2)

  def action_button_overlay(self):
    """ Toggle the overlay visibility
    """
    if(self.overlay==0):
      self.overlay=1
    else:
      self.overlay=0

  def action_button_zoom(self):
    """ Toggle the zoom window visibility
    """
    #print("dbg656: action_button_zoom")
    #print("dbg567: frame_b.winfo_exists:", self.frame_b.winfo_exists())
    #print("dbg568: frame_b.state:", self.frame_b.state())
    if(self.frame_b.state()=='withdrawn'):
      self.show_zoom_frame()
    else:
      self.hide_zoom_frame()

  def create_zoom_frame(self):
    """ Define the zoom window
    """
    #print("dbg554: create_zoom_frame")
    self.frame_b = Tkinter.Toplevel(self.frame_a)
    #self.frame_b.grid(column=0, row=0, sticky=Tkinter.N+Tkinter.E+Tkinter.S+Tkinter.W) # Toplevel doesn't have grid method !
    #self.frame_b.pack(fill=Tkinter.BOTH, expand=1)
    self.frame_b.title("cnc25d display backend details")
    self.canvas_b = Tkinter.Canvas(self.frame_b, width=initial_tkinter_canvas_width, height=initial_tkinter_canvas_height)
    #self.canvas_b.grid(column=0, row=0, sticky=Tkinter.N+Tkinter.E+Tkinter.S+Tkinter.W)
    #self.canvas_b.columnconfigure(0, weight=1)
    #self.canvas_b.rowconfigure(0, weight=1)
    self.canvas_b.pack(fill=Tkinter.BOTH, expand=1) # with Toplevel parent, it seems you need to use pack to resisze the canvas !
    self.canvas_b.bind("<ButtonPress-1>", self.action_canvas_b_mouse_button_press)
    self.canvas_b.bind("<B1-Motion>", self.action_canvas_b_mouse_button_motion)
    self.canvas_b.bind("<ButtonRelease-1>", self.action_canvas_b_mouse_button_release)
    self.frame_b.protocol("WM_DELETE_WINDOW", self.hide_zoom_frame) # change the behaviour of the window X button

  def hide_zoom_frame(self):
    """ Hide the zoom window
    """
    self.frame_b.withdraw()

  def show_zoom_frame(self):
    """ Show the zoom window
    """
    self.frame_b.update()
    self.frame_b.deiconify()

  def action_button_parameters(self):
    """ Toggle the parameter window visibility
    """
    #tkMessageBox.showinfo('lala','Yes mes')
    #print("dbg566: frame_c.winfo_exists:", self.frame_c.winfo_exists())
    #print("dbg569: frame_c.state:", self.frame_c.state())
    #print("dbg559: self.parameter_content.get():", self.parameter_content.get())
    if(self.frame_c.state()=='withdrawn'):
      if(self.parameter_content.get()==''):
        print("WARN556: Warning, no parameter information has been set!")
      else:
        self.show_parameter_frame()
    else:
      self.hide_parameter_frame()


  def create_parameter_frame(self):
    """ Define the parameter window
    """
    #print("dbg414: create_parameter_frame")
    self.frame_c = Tkinter.Toplevel(self.frame_a)
    self.frame_c.title("parameter info")
    self.parameter_message = Tkinter.Message(self.frame_c, textvariable=self.parameter_content)
    self.parameter_message.grid(sticky=Tkinter.N+Tkinter.E+Tkinter.S+Tkinter.W)
    self.frame_c.protocol("WM_DELETE_WINDOW", self.hide_parameter_frame) # change the behaviour of the window X button

  def hide_parameter_frame(self):
    """ Hide the parameter window
    """
    self.frame_c.withdraw()

  def show_parameter_frame(self):
    """ Show the parameter window
    """
    self.frame_c.update()
    self.frame_c.deiconify()

  #def apply_curve_graphic_table(self, ai_angle_position, ai_angle_speed):
  #  """ Compute the curve new points accordind to the angle_position and angle_speed
  #  """
  #  curve_nb = len(self.curve_graphic_table)-1
  #  if(curve_nb>0):
  #    # angle in x-axis
  #    #lx = ai_angle_position
  #    # time in x-axis
  #    lx = 0 # initial timestamp
  #    if(len(self.curve_points[0])>0):
  #      #dlx = ai_angle_speed/self.curve_graphic_table[0][2]
  #      dlx = self.curve_graphic_table[0][2]
  #      lx = self.curve_points[0][-1] + abs(dlx)
  #    self.curve_points[0].append(lx)
  #    for i in range(curve_nb):
  #      ly = self.curve_graphic_table[i+1][1](ai_angle_position, ai_angle_speed)
  #      self.curve_points[i+1].append(ly)

  def action_button_curve_graph(self, event):
    """ Launch the matplotlib window for curve display
    """
    curve_nb = len(self.curve_graphic_table)-1
    if(curve_nb<=0):
      print("WARN451: Warning, self.curve_graphic_table is not set!")
    else:
      # create curve_points
      curve_nb = len(self.curve_graphic_table)-1
      curve_points = []
      for i in range(curve_nb+1):
        curve_points.append([])
      curve_table_len = len( self.curve_graphic_table[1][1])
      for j in range(curve_nb-1):
        check_len = len( self.curve_graphic_table[j+2][1])
        if(check_len!=curve_table_len):
          print("ERR586: Error in the curve_table {:d}! Its lenght {:d} does not match the reference length {:d}".format(j, check_len, curve_table_len))
      x_increment = self.curve_graphic_table[0][2]
      abscissa_x = 0
      for i in range(curve_table_len):
        curve_points[0].append(abscissa_x)
        abscissa_x += x_increment
        for j in range(curve_nb):
          table = self.curve_graphic_table[j+1][1]
          ly = table[i]
          curve_points[j+1].append(ly)
      #matplotlib.pyplot.plot([1,2,3,4,5],[5,3,2,1,4])
      matplotlib.pyplot.figure(1)
      for i in range(curve_nb):
        matplotlib.pyplot.subplot(curve_nb,1,i+1)
        #matplotlib.pyplot.plot(self.curve_points[0], self.curve_points[i+1], self.curve_graphic_table[i+1][2])
        matplotlib.pyplot.plot(curve_points[0], curve_points[i+1], self.curve_graphic_table[i+1][2])
        matplotlib.pyplot.ylabel(self.curve_graphic_table[i+1][0])
        if(i==0):
          matplotlib.pyplot.title(self.curve_graphic_table[0][0])
        if(i==curve_nb-1):
          matplotlib.pyplot.xlabel(self.curve_graphic_table[0][1])
      matplotlib.pyplot.show()

  def createWidgets(self):
    """ Create the widgets of the main window with their layout and also the zoom window and parameter window
    """
    #
    #self.frame_canvas = Tkinter.Frame(self.frame_a)
    ##self.frame_canvas.pack(side=Tkinter.TOP)
    #self.frame_canvas.grid(column=0, row=0, sticky=Tkinter.W + Tkinter.N + Tkinter.E + Tkinter.S)
    #self.frame_canvas.columnconfigure(0, weight=1)
    #self.frame_canvas.rowconfigure(0, weight=1)
    #
    #self.canvas_a =  Tkinter.Canvas(self.frame_a, width=self.canvas_a_width, height=self.canvas_a_height)
    self.canvas_a =  Tkinter.Canvas(self.frame_a, width=initial_tkinter_canvas_width, height=initial_tkinter_canvas_height)
    #self.canvas_a.pack(fill=Tkinter.BOTH, expand=1)
    self.canvas_a.grid(column=0, row=0, sticky=Tkinter.N+Tkinter.E+Tkinter.S+Tkinter.W)
    self.canvas_a.columnconfigure(0, weight=1)
    self.canvas_a.rowconfigure(0, weight=1)
    self.canvas_a.bind("<ButtonPress-1>", self.action_canvas_a_mouse_button_press)
    self.canvas_a.bind("<B1-Motion>", self.action_canvas_a_mouse_button_motion)
    self.canvas_a.bind("<ButtonRelease-1>", self.action_canvas_a_mouse_button_release)
    #
    self.frame_control = Tkinter.Frame(self.frame_a)
    #self.frame_control.pack(side=Tkinter.BOTTOM)
    self.frame_control.grid(column=0, row=1, sticky=Tkinter.W + Tkinter.S)
    #
    self.frame_button_speed = Tkinter.Frame(self.frame_control)
    self.frame_button_speed.grid(column=0, row=0)
    #
    self.button_fast_backward = Tkinter.Button(self.frame_button_speed)
    self.button_fast_backward["text"] = "<<-",
    self.button_fast_backward["command"] = self.action_button_fast_backward
    self.button_fast_backward.pack(side=Tkinter.LEFT)
    #
    self.button_slow_backward = Tkinter.Button(self.frame_button_speed)
    self.button_slow_backward["text"] = "<-",
    self.button_slow_backward["command"] = self.action_button_slow_backward
    self.button_slow_backward.pack(side=Tkinter.LEFT)
    #
    self.button_step_backward = Tkinter.Button(self.frame_button_speed)
    self.button_step_backward["text"] = "<|",
    self.button_step_backward["command"] = self.action_button_step_backward
    self.button_step_backward.pack(side=Tkinter.LEFT)
    #
    self.button_stop = Tkinter.Button(self.frame_button_speed)
    self.button_stop["foreground"]   = "red"
    self.button_stop["text"] = "||",
    self.button_stop["command"] = self.action_button_stop
    self.button_stop.pack(side=Tkinter.LEFT)
    #
    self.button_step_forward = Tkinter.Button(self.frame_button_speed)
    self.button_step_forward["text"] = "|>",
    self.button_step_forward["command"] = self.action_button_step_forward
    self.button_step_forward.pack(side=Tkinter.LEFT)
    #
    self.button_slow_forward = Tkinter.Button(self.frame_button_speed)
    self.button_slow_forward["text"] = "->",
    self.button_slow_forward["command"] = self.action_button_slow_forward
    self.button_slow_forward.pack(side=Tkinter.LEFT)
    #
    self.button_fast_forward = Tkinter.Button(self.frame_button_speed)
    self.button_fast_forward["text"] = "->>",
    self.button_fast_forward["command"] = self.action_button_fast_forward
    self.button_fast_forward.pack(side=Tkinter.LEFT)
    #
    self.frame_label = Tkinter.Frame(self.frame_control)
    self.frame_label.grid(column=0, row=1, sticky=Tkinter.W)
    #
    self.label_angle = Tkinter.Label(self.frame_label)
    self.label_angle["textvariable"] = self.label1_content
    #self.label_angle.pack(side=Tkinter.TOP)
    #self.label_angle.pack(anchor=Tkinter.NW)
    self.label_angle.grid(column=0, row=0, sticky=Tkinter.W)
    #
    self.label_angle_speed = Tkinter.Label(self.frame_label)
    self.label_angle_speed["textvariable"] = self.label2_content
    #self.label_angle_speed.pack(side=Tkinter.BOTTOM)
    #self.label_angle_speed.pack(anchor=Tkinter.SW)
    self.label_angle_speed.grid(column=0, row=1, sticky=Tkinter.W)
    #
    self.frame_button_options = Tkinter.Frame(self.frame_control)
    self.frame_button_options.grid(column=0, row=2, sticky=Tkinter.W+Tkinter.S)
    #
    self.button_zoom = Tkinter.Button(self.frame_button_options)
    self.button_zoom["text"] = "Zoom",
    self.button_zoom["command"] = self.action_button_zoom
    self.button_zoom.pack(side=Tkinter.LEFT)
    #
    self.button_overlay = Tkinter.Button(self.frame_button_options)
    self.button_overlay["text"] = "Overlay",
    self.button_overlay["command"] = self.action_button_overlay
    self.button_overlay.pack(side=Tkinter.LEFT)
    #
    self.button_parameter = Tkinter.Button(self.frame_button_options)
    self.button_parameter["text"] = "Parameters",
    self.button_parameter["command"] = self.action_button_parameters
    self.button_parameter.pack(side=Tkinter.LEFT)
    #
    self.button_graph = Tkinter.Button(self.frame_button_options)
    self.button_graph["text"] = "Graph",
    #self.button_graph["command"] = lambda a=1: self.action_button_check_event(a)
    #self.button_graph.bind("<Button-1>", lambda event: self.action_button_check_event(event))
    #self.button_graph.bind("<Button-1>", self.action_button_check_event) # the event argument is added by default
    self.button_graph.bind("<Button-1>", self.action_button_curve_graph) # the event argument is added by default
    self.button_graph.pack(side=Tkinter.LEFT)
    #
    self.button_quit = Tkinter.Button(self.frame_button_options)
    self.button_quit["text"] = "Quit",
    #self.button_quit["command"] = self.frame_a.quit
    self.button_quit["command"] = self.tktop.destroy
    self.button_quit.pack(side=Tkinter.LEFT)
    #
    ## second window with canvas_b
    self.create_zoom_frame()
    self.show_zoom_frame()
    #
    ## third window with parameter text
    self.create_parameter_frame()
    self.hide_parameter_frame()

  def __init__(self, winParent):
    """ Initiate the class Two_Canvas by creating the windows, initializing the class variables and starting the time simulation
    """
    self.tktop = winParent # used to destroy the app when quit
    self.frame_a = Tkinter.Frame(winParent)
    winParent.title("cnc25d display backend main")
    #self.frame_a.pack()
    #winParent.grid(sticky=Tkinter.N+Tkinter.E+Tkinter.S+Tkinter.W)
    winParent.columnconfigure(0, weight=1)
    winParent.rowconfigure(0, weight=1)
    self.frame_a.grid(sticky=Tkinter.N+Tkinter.E+Tkinter.S+Tkinter.W)
    self.frame_a.columnconfigure(0, weight=1)
    self.frame_a.rowconfigure(0, weight=1)
    #
    self.angle_speed = 0*g_slow_angle_speed
    self.rotation_direction = 1
    self.angle_position = 0
    self.outline_extremum = (0,0,0,0)
    #
    self.label1_content = Tkinter.StringVar()
    self.label2_content = Tkinter.StringVar()
    self.set_label_content()
    #
    self.parameter_content = Tkinter.StringVar()
    #
    self.overlay = 0
    self.canvas_graphic_function = None
    self.mouse_x1 = 0
    self.mouse_y1 = 0
    self.mouse_x2 = initial_tkinter_canvas_width/2
    self.mouse_y2 = initial_tkinter_canvas_height/2
    self.canvas_a_mouse_press = 0
    self.scale_coef_a = None
    self.crop_limit = (0,0,0,0)
    self.mouse_bx1 = 0
    self.mouse_by1 = 0
    self.mouse_bx2 = 0
    self.mouse_by2 = 0
    self.canvas_b_mouse_press = 0
    self.measurement_id = 0
    self.scale_coef_b = None
    #
    self.curve_graphic_table = []
    #self.curve_points = []
    #
    self.createWidgets()
    # initiate the time simulation
    self.simulation_step()

  def add_canvas_graphic_function(self, ai_canvas_graphic_function):
    """ api method to add or change the canvas graphics
    """
    #print("dbg445: ai_graphic_function:", ai_graphic_function)
    self.canvas_graphic_function = ai_canvas_graphic_function
    # the next line is to compute the extremum once at initialization if you don't want to recompute it at each angle_position
    self.outline_extremum = find_outline_extremum(self.canvas_graphic_function(self.rotation_direction, self.angle_position))

  def add_parameter_info(self, ai_parameter_info):
    """ api method to add or change the parameter info
    """
    self.parameter_content.set(ai_parameter_info)
    
  def add_curve_graphic_table(self, ai_curve_graphic_table):
    """ api method to add or change the curve functions for matplotlib
    """
    self.curve_graphic_table = ai_curve_graphic_table
    #curve_nb = len(self.curve_graphic_table)-1
    #self.curve_points = []
    #for i in range(curve_nb+1):
    #  self.curve_points.append([])



#  tk_a = Tkinter.Tk()
#  tk_a.title("test1 backend tkinter")
#  canvas_a = Tkinter.Canvas(tk_a, width=win_a_w,height=win_a_h)
#  canvas_a.bind("<Configure>", gg_redraw_canvas_a)
#  canvas_a.bind("<Button-1>", gg_callback_on_click_a)
#  canvas_a.pack(fill=Tkinter.BOTH, expand=1)
#  tk_a.mainloop()

################################################################
# ******** test the Two_Canvas class ***********
################################################################

def test_canvas_graphic_1(ai_rotation_direction, ai_angle):
  """ Sub-function for test 1
      Also example of the callback function for add_canvas_graphic_function
  """
  polygon_test_1 = (
    5,5,
    35,5,
    20,35)
  lines_test_1 = (
    (0,0,20,25),
    (5,5,20,25))
  polygon_test_2 = (
    105,105,
    135,105,
    120,135)
  lines_test_2 = (
    (15,0,50,15),
    (-30,30,50,15))
  polygon_test_3 = (
    205,205,
    235,205,
    220,265)
  r_canvas_graph = []
  r_canvas_graph.append(('graphic_lines', lines_test_1, 'red', 1))
  r_canvas_graph.append(('graphic_polygon', polygon_test_1, 'green', 'red', 1))
  r_canvas_graph.append(('overlay_lines', lines_test_2, 'yellow', 2))
  r_canvas_graph.append(('overlay_polygon', polygon_test_2, 'orange', 'yellow', 2))
  r_canvas_graph.append(('graphic_polygon', polygon_test_3, 'green', 'red', 1))
  return(r_canvas_graph)

#def test1_curve1(ai_angle_position, ai_angle_speed):
#  """ test curve for matplotlib
#  """
#  r_y = math.sin(ai_angle_position)
#  return(r_y)
#
#def test1_curve2(ai_angle_position, ai_angle_speed):
#  """ test curve for matplotlib
#  """
#  r_y = math.cos(ai_angle_position)
#  return(r_y)

def two_canvas_class_test1():
  """ test the simple display of a static graphic with Two_Canvas
  """
  #
  test_parameter_info = """
Maître Corbeau, sur un arbre perché,
Tenait en son bec un fromage.
Maître Renard, par l'odeur alléché,
Lui tint à peu près ce langage :
"Hé ! bonjour, Monsieur du Corbeau.
Que vous êtes joli ! que vous me semblez beau !
Sans mentir, si votre ramage
Se rapporte à votre plumage,
Vous êtes le Phénix des hôtes de ces bois. "
  """
  #
  #lambda_function_1 = test1_curve1
  #lambda_function_1 = lambda x: test1_curve1(x)
  #lambda_function_2 = test1_curve2
  test1_curve1 = (1,2,3,4,5,6,7,8,9,10,11,12,13)
  test1_curve2 = (8,7,3,4,2,2,3,4,5,5.5,6,6.2,6.9)
  test_curve_graphic_1 = (('global_title', 'x_axis_name', 1),
    ('plot1_title', test1_curve1, 'bo'),
    ('plot2_title', test1_curve2, 'r'))
  #
  tk_root = Tkinter.Tk()
  dut = Two_Canvas(tk_root)
  #dut = Two_Canvas()
  dut.add_canvas_graphic_function(test_canvas_graphic_1)
  dut.add_parameter_info(test_parameter_info)
  dut.add_curve_graphic_table(test_curve_graphic_1)
  tk_root.mainloop()
  #dut.mainloop()
  r_test = 1
  return(r_test)

def test_canvas_graphic_2(ai_rotation_direction, ai_angle):
  """ Sub-function for test 1
      Also example of the callback function for add_canvas_graphic_function
  """
  l_angle = float(ai_angle)/100
  polygon_test_2 = (
    30*math.cos(l_angle+1.0*math.pi/2), 30*math.sin(l_angle+1.0*math.pi/2),
    35*math.cos(l_angle+1.5*math.pi/2), 35*math.sin(l_angle+1.5*math.pi/2),
    20*math.cos(l_angle+2.0*math.pi/2), 20*math.sin(l_angle+2.0*math.pi/2),
    50*math.cos(l_angle+2.5*math.pi/2), 50*math.sin(l_angle+2.5*math.pi/2),
    40*math.cos(l_angle+3.0*math.pi/2), 40*math.sin(l_angle+3.0*math.pi/2),
    30*math.cos(l_angle+3.5*math.pi/2), 30*math.sin(l_angle+3.5*math.pi/2),
    25*math.cos(l_angle+4.0*math.pi/2), 25*math.sin(l_angle+4.0*math.pi/2),
    35*math.cos(l_angle+4.5*math.pi/2), 35*math.sin(l_angle+4.5*math.pi/2))
  r_canvas_graph = []
  r_canvas_graph.append(('graphic_polygon', polygon_test_2, '', 'red', 1))
  return(r_canvas_graph)

def two_canvas_class_test2():
  """ test the simple display of a dynamic graphic with Two_Canvas
  """
  #
  test_parameter_info = """
Maître Corbeau, sur un arbre perché,
Tenait en son bec un fromage.
  """
  #
  test1_curve1 = (1,2,3,4,5,6,7,8,9,10,11,12,13)
  test1_curve2 = (8,7,3,4,2,2,3,4,5,5.5,6,6.2,6.9)
  test_curve_graphic_1 = (('global_title', 'x_axis_name', 0.001),
    ('plot1_title', test1_curve1, 'bo'),
    ('plot2_title', test1_curve2, 'r'))
  #
  tk_root = Tkinter.Tk()
  dut = Two_Canvas(tk_root)
  dut.add_canvas_graphic_function(test_canvas_graphic_2)
  dut.add_parameter_info(test_parameter_info)
  dut.add_curve_graphic_table(test_curve_graphic_1)
  tk_root.mainloop()
  #
  r_test = 1
  return(r_test)

################################################################
# ******** command line interface ***********
################################################################

def display_backends_cli(ai_args=None):
  """ command line interface to run this script in standalone
  """
  db_parser = argparse.ArgumentParser(description='Test the display_backend Two_Canvas).')
  db_parser.add_argument('--test1','--t1', action='store_true', default=False, dest='sw_test1',
    help='Run two_canvas_class_test1() with a static graphic')
  db_parser.add_argument('--test2','--t2', action='store_true', default=False, dest='sw_test2',
    help='Run two_canvas_class_test2() with a dynamic graphic')
  effective_args = design_help.get_effective_args(ai_args)
  db_args = db_parser.parse_args(effective_args)
  r_dbc = 0
  print("dbg111: start testing display_backends.py")
  if(db_args.sw_test1):
    r_dbc = two_canvas_class_test1()
  elif(db_args.sw_test2):
    r_dbc = two_canvas_class_test2()
  print("dbg999: end of script")
  return(r_dbc)

################################################################
# main
################################################################

if __name__ == "__main__":
  print("display_backend.py says hello!")
  # choose the script behavior
  #display_backends_cli()                   # get arguments from the command line
  display_backends_cli("--test1".split())   # run the test1
  #display_backends_cli("--test2".split())




# gear_profile_macro.py
# the macro to generate a gear_profile.
# created by charlyoleg on 2013/08/27
#
# (C) Copyright 2013 charlyoleg
#
# This file is part of the Cnc25D Python package.
# 
# Cnc25D is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Cnc25D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Cnc25D.  If not, see <http://www.gnu.org/licenses/>.


################################################################
# this file intends being included in the file bin/cnc25d_example_generator.py
# for this purpose, there is some syntaxe restrictions
# don't use triple single-quotes (') and return character ('\'.'n') in this file
# but you can still use triple double-quote (")
################################################################

"""
this piece of code is an example of how to use the parametric design gear_profile
You can also use this file as a FreeCAD macro from the GUI
Look at the code. It's very simple to hack
"""

################################################################
# Installation pre-request
################################################################
# This script needs freecad and Cnc25D installed on your system
# visit those sites for more information:
# http://www.freecadweb.org/
# https://pypi.python.org/pypi/Cnc25D
#
# To install FreeCAD on Ubuntu, run the following command:
# > sudo apt-get install freecad
# or to get the newest version:
# > sudo add-apt-repository ppa:freecad-maintainers/freecad-stable
# > sudo apt-get update
# > sudo apt-get install freecad
# and optionally:
# >  sudo apt-get install freecad-doc freecad-dev
# To install the python package cnc25d, run the following command:
# > sudo pip install Cnc25D
# or
# > sudo pip install Cnc25D -U


################################################################
# header for Python / FreeCAD compatibility
################################################################

from cnc25d import cnc25d_api
cnc25d_api.importing_freecad()

#print("FreeCAD.Version:", FreeCAD.Version())

################################################################
# import
################################################################

#
import math
#
from cnc25d import gear_profile
#
import Part

    
################################################################
# parameters value
################################################################
#
# choose the values of the parameters by editing this file
# feature request : create a GUI with PyQt4 to edit those parameter values

### first gear
# general
gp_gear_type                      = 'e'
gp_gear_tooth_nb                  = 18
gp_gear_module                    = 10.0
gp_gear_primitive_diameter        = 0
gp_gear_addendum_dedendum_parity  = 50.0
# tooth height
gp_gear_tooth_half_height           = 0 # 10.0
gp_gear_addendum_height_pourcentage = 100.0
gp_gear_dedendum_height_pourcentage = 100.0
gp_gear_hollow_height_pourcentage   = 25.0
gp_gear_router_bit_radius           = 3.0
# positive involute
gp_gear_base_diameter       = 0
gp_gear_force_angle         = 0
gp_gear_tooth_resolution    = 2
gp_gear_skin_thickness      = 0
# negative involute (if zero, negative involute = positive involute)
gp_gear_base_diameter_n     = 0
gp_gear_force_angle_n       = 0
gp_gear_tooth_resolution_n  = 0
gp_gear_skin_thickness_n    = 0
### second gear
# general
gp_second_gear_type                     = 'e'
gp_second_gear_tooth_nb                 = 0 #25
gp_second_gear_primitive_diameter       = 0
gp_second_gear_addendum_dedendum_parity = 0 # 50.0
# tooth height
gp_second_gear_tooth_half_height            = 0
gp_second_gear_addendum_height_pourcentage  = 100.0
gp_second_gear_dedendum_height_pourcentage  = 100.0
gp_second_gear_hollow_height_pourcentage    = 25.0
gp_second_gear_router_bit_radius            = 0
# positive involute
gp_second_gear_base_diameter      = 0
gp_second_gear_tooth_resolution   = 0
gp_second_gear_skin_thickness     = 0
# negative involute (if zero, negative involute = positive involute)
gp_second_gear_base_diameter_n    = 0
gp_second_gear_tooth_resolution_n = 0
gp_second_gear_skin_thickness_n   = 0
### gearbar specific
gp_gearbar_slope                  = 0.0
gp_gearbar_slope_n                = 0.0
### position
# first gear position
gp_center_position_x                    = 0.0
gp_center_position_y                    = 0.0
gp_gear_initial_angle                   = 0.0
# second gear position
gp_second_gear_position_angle           = 0.0
gp_second_gear_additional_axis_length   = 0.0
### portion
gp_portion_tooth_nb     = 10
gp_portion_first_end    = 3
gp_portion_last_end     = 3
### output1
gp_gear_profile_height  = 20.0
gp_simulation_enable    = True
gp_output_file_basename = "" # set a not-empty string if you want to generate the output files
#gp_output_file_basename = "test_output/gearwheel_macro.svg"  # to generate the SVG file with mozman svgwrite
#gp_output_file_basename = "test_output/gearwheel_macro.dxf"  # to generate the DXF file with mozman svgwrite
#gp_output_file_basename = "test_output/gearwheel_macro"      # to generate the Brep and DXF file with FreeCAD



################################################################
# action
################################################################

my_gp = gear_profile.gear_profile(
            ### first gear
            # general
            ai_gear_type                      = gp_gear_type,
            ai_gear_tooth_nb                  = gp_gear_tooth_nb,
            ai_gear_module                    = gp_gear_module,
            ai_gear_primitive_diameter        = gp_gear_primitive_diameter,
            ai_gear_addendum_dedendum_parity  = gp_gear_addendum_dedendum_parity,
            # tooth height
            ai_gear_tooth_half_height           = gp_gear_tooth_half_height,
            ai_gear_addendum_height_pourcentage = gp_gear_addendum_height_pourcentage,
            ai_gear_dedendum_height_pourcentage = gp_gear_dedendum_height_pourcentage,
            ai_gear_hollow_height_pourcentage   = gp_gear_hollow_height_pourcentage,
            ai_gear_router_bit_radius           = gp_gear_router_bit_radius,
            # positive involute
            ai_gear_base_diameter       = gp_gear_base_diameter,
            ai_gear_force_angle         = gp_gear_force_angle,
            ai_gear_tooth_resolution    = gp_gear_tooth_resolution,
            ai_gear_skin_thickness      = gp_gear_skin_thickness,
            # negative involute (if zero, negative involute = positive involute)
            ai_gear_base_diameter_n     = gp_gear_base_diameter_n,
            ai_gear_force_angle_n       = gp_gear_force_angle_n,
            ai_gear_tooth_resolution_n  = gp_gear_tooth_resolution_n,
            ai_gear_skin_thickness_n    = gp_gear_skin_thickness_n,
            ### second gear
            # general
            ai_second_gear_type                     = gp_second_gear_type,
            ai_second_gear_tooth_nb                 = gp_second_gear_tooth_nb,
            ai_second_gear_primitive_diameter       = gp_second_gear_primitive_diameter,
            ai_second_gear_addendum_dedendum_parity = gp_second_gear_addendum_dedendum_parity,
            # tooth height
            ai_second_gear_tooth_half_height            = gp_second_gear_tooth_half_height,
            ai_second_gear_addendum_height_pourcentage  = gp_second_gear_addendum_height_pourcentage,
            ai_second_gear_dedendum_height_pourcentage  = gp_second_gear_dedendum_height_pourcentage,
            ai_second_gear_hollow_height_pourcentage    = gp_second_gear_hollow_height_pourcentage,
            ai_second_gear_router_bit_radius            = gp_second_gear_router_bit_radius,
            # positive involute
            ai_second_gear_base_diameter      = gp_second_gear_base_diameter,
            ai_second_gear_tooth_resolution   = gp_second_gear_tooth_resolution,
            ai_second_gear_skin_thickness     = gp_second_gear_skin_thickness,
            # negative involute (if zero, negative involute = positive involute)
            ai_second_gear_base_diameter_n    = gp_second_gear_base_diameter_n,
            ai_second_gear_tooth_resolution_n = gp_second_gear_tooth_resolution_n,
            ai_second_gear_skin_thickness_n   = gp_second_gear_skin_thickness_n,
            ### gearbar specific
            ai_gearbar_slope                  = gp_gearbar_slope,
            ai_gearbar_slope_n                = gp_gearbar_slope_n,
            ### position
            # first gear position
            ai_center_position_x                    = gp_center_position_x,
            ai_center_position_y                    = gp_center_position_y,
            ai_gear_initial_angle                   = gp_gear_initial_angle,
            # second gear position
            ai_second_gear_position_angle           = gp_second_gear_position_angle,
            ai_second_gear_additional_axis_length   = gp_second_gear_additional_axis_length,
            ### portion
            ai_portion_tooth_nb     = gp_portion_tooth_nb,
            ai_portion_first_end    = gp_portion_first_end,
            ai_portion_last_end     = gp_portion_last_end,
            ### output
            ai_gear_profile_height  = gp_gear_profile_height,
            ai_simulation_enable    = gp_simulation_enable,    # gp_simulation_enable,
            ai_output_file_basename = gp_output_file_basename)

#print("dbg339: my_gp:", my_gp)
#Part.show(my_gp)





# design_help.py
# small help functions for design scripts
# created by charlyoleg on 2013/08/27
#
# (C) Copyright 2013 charlyoleg
#
# This file is part of the Cnc25D Python package.
# 
# Cnc25D is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Cnc25D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Cnc25D.  If not, see <http://www.gnu.org/licenses/>.


"""
design_help.py is part of the Cnc25D API.
it provides macros that are commonly used in design scripts
"""

################################################################
# header for Python / FreeCAD compatibility
################################################################

#import importing_freecad
#importing_freecad.importing_freecad()

#print("FreeCAD.Version:", FreeCAD.Version())
#FreeCAD.Console.PrintMessage("Hello from PrintMessage!\n") # avoid using this method because it is not printed in the FreeCAD GUI

################################################################
# import
################################################################

# Python standard library
#import math
import sys, argparse
#from datetime import datetime
import os, errno
#import re
#import Tkinter # to display the outline in a small GUI
# FreeCAD
#import Part
#from FreeCAD import Base
# 3rd parties
#import svgwrite
#from dxfwrite import DXFEngine
# cnc25d
#import outline_backends
#import export_2d


################################################################
# help functions
################################################################

################################################################
# functions to be reused
################################################################

def mkdir_p(ai_directory):
  """ Create the directory if needed and possible
  """
  r_status = 2
  if(ai_directory!=''):
    #print("dbg448: ai_directory:", ai_directory)
    # mkdir -p directory if needed
    print("dbg450: try to create the output directory: {:s}".format(ai_directory))
    try:
      os.makedirs(ai_directory)
      r_status = 0
    except OSError as exc:
      if exc.errno == errno.EEXIST and os.path.isdir(ai_directory):
        pass
      else:
        raise
  return(r_status)

def get_effective_args(ai_default_args):
  """ Find the args to be processed by the parser.
      Use sys.argv, but if empty then use ai_default_args
      let run your script with python and freecad
  """
  # this ensure the possible to use the script with python and freecad
  # You can not use argparse and FreeCAD together, so it's actually useless !
  # Running this script, FreeCAD will just use the argparse default values
  arg_index_offset=0
  if(sys.argv[0]=='freecad'): # check if the script is used by freecad
    arg_index_offset=1
    if(len(sys.argv)>=2):
      if(sys.argv[1]=='-c'): # check if the script is used by freecad -c
        arg_index_offset=2
  r_effective_args = sys.argv[arg_index_offset+1:]
  if(len(r_effective_args)==0):
    r_effective_args = ai_default_args
  #print("dbg115: r_effective_args:", str(r_effective_args))
  #FreeCAD.Console.PrintMessage("dbg116: r_effective_args: %s\n"%(str(r_effective_args)))
  return(r_effective_args)

################################################################
# test-functions
################################################################





# small_geometry.py
# a help module for cnc_cut_outline.py, just to avoid to get too big files
# created by charlyoleg on 2013/07/30
#
# (C) Copyright 2013 charlyoleg
#
# This file is part of the Cnc25D Python package.
# 
# Cnc25D is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Cnc25D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Cnc25D.  If not, see <http://www.gnu.org/licenses/>.


"""
small_geometry.py provides sub functions for cnc_cut_outline.py related to the Euclid geometry
"""

################################################################
# python behavior
################################################################

from __future__ import division # to get float division


################################################################
# import
################################################################

#import outline_backends # for testing only
#
import math
import sys, argparse

################################################################
# functions to be used by cnc_cut_outline.py
################################################################

def rotate_point(ai_point, ai_ox, ai_oy, ai_rotation_angle):
  """ Rotation of the point ai_point of center (ai_ox, ai_oy) and angle ai_rotation_angle
  """
  ix = ai_point[0]-ai_ox
  iy = ai_point[1]-ai_oy
  pt_x = ai_ox+ix*math.cos(ai_rotation_angle)-iy*math.sin(ai_rotation_angle)
  pt_y = ai_oy+ix*math.sin(ai_rotation_angle)+iy*math.cos(ai_rotation_angle)
  r_point = [pt_x, pt_y]
  return(r_point)

def arc_center_radius(ai_arc_pt1, ai_arc_pt2, ai_arc_pt3, ai_error_msg_id):
  """ Compute the center of the arc defined by the three points A,B and C
  """
  #print("dbg197: ai_error_msg_id: {:s}".format(ai_error_msg_id))
  # use to check is angle is smaller than pi/2
  radian_epsilon = math.pi/1000
  # interpretation of the arc three points A,B,C
  AX = ai_arc_pt1[0]
  AY = ai_arc_pt1[1]
  BX = ai_arc_pt2[0]
  BY = ai_arc_pt2[1]
  CX = ai_arc_pt3[0]
  CY = ai_arc_pt3[1]
  # check of the lenght AB, BC, AC
  AB = math.sqrt((BX-AX)**2+(BY-AY)**2)
  BC = math.sqrt((CX-BX)**2+(CY-BY)**2)
  AC = math.sqrt((CX-AX)**2+(CY-AY)**2)
  if((AB<radian_epsilon)or(BC<radian_epsilon)or(AC<radian_epsilon)):
    print("ERR682: Error in {:s}, the three arc points ABC are too closed: AB={:0.2f} BC={:0.2f} AC={:0.2f}".format(ai_error_msg_id, AB, BC, AC))
    sys.exit(2)
  # calculation of M and N
  MX = (AX+BX)/2
  MY = (AY+BY)/2
  NX = (BX+CX)/2
  NY = (BY+CY)/2
  # calculation of e and f
  cos_e = (BX-AX)/AB
  sin_e = (BY-AY)/AB
  cos_f = (CX-BX)/BC
  sin_f = (CY-BY)/BC
  # calculation de I
  #(cos(e)*sin(f)-cos(f)*sin(e))*x = sin(f)*(cos(e)*xM+sin(e)*yM)-sin(e)*(cos(f)*xN+sin(f)*yN)
  #(sin(e)*cos(f)-sin(f)*cos(e))*y = cos(f)*(cos(e)*xM+sin(e)*yM)-cos(e)*(cos(f)*xN+sin(f)*yN)
  # ixl = (cos(e)*sin(f)-cos(f)*sin(e))
  # iyl = (sin(e)*cos(f)-sin(f)*cos(e))
  # ixk = sin(f)*(cos(e)*xM+sin(e)*yM)-sin(e)*(cos(f)*xN+sin(f)*yN)
  # iyk = cos(f)*(cos(e)*xM+sin(e)*yM)-cos(e)*(cos(f)*xN+sin(f)*yN)
  ixl = cos_e*sin_f-cos_f*sin_e
  iyl = sin_e*cos_f-sin_f*cos_e
  ixk = sin_f*(cos_e*MX+sin_e*MY)-sin_e*(cos_f*NX+sin_f*NY)
  iyk = cos_f*(cos_e*MX+sin_e*MY)-cos_e*(cos_f*NX+sin_f*NY)
  if((abs(ixl)<radian_epsilon)or(abs(iyl)<radian_epsilon)):
    print("ERR947: Error in {:s}, ixl (= {:0.2f}) or iyl (= {:0.2f}) are too closed to zero!".format(ai_error_msg_id, ixl, iyl))
    sys.exit(2)
  IX=ixk/ixl
  IY=iyk/iyl
  # check than I is equidistant of A, B and C
  IA = math.sqrt((AX-IX)**2+(AY-IY)**2)
  IB = math.sqrt((BX-IX)**2+(BY-IY)**2)
  IC = math.sqrt((CX-IX)**2+(CY-IY)**2)
  if((abs(IB-IA)>radian_epsilon)or(abs(IC-IA)>radian_epsilon)):
    print("ERR748: Error in {:s}, the calculation of the center of the arc A,B,C is wrong! IA={:0.2f} IB={:0.2f} IC={:0.2f}".format(ai_error_msg_id, IA, IB, IC))
    print("dbg253: A= {:0.2f} {:0.2f}  B= {:0.2f} {:0.2f}  C= {:0.2f} {:0.2f}  I= {:0.2f} {:0.2f}".format(AX,AY,BX,BY,CX,CY,IX,IY))
    print("dbg764: cos_e={:0.2f}  sin_e={:0.2f}  cos_f={:0.2f}  sin_f={:0.2f}".format(cos_e, sin_e, cos_f, sin_f))
    print("dbg765: ixl={:0.2f} ixk={:0.2f} iyl={:0.2f} iyk={:0.2f}".format(ixl, ixk, iyl, iyk))
    print("dbg766: MX={:0.2f} MY={:0.2f} NX={:0.2f} NY={:0.2f}".format(MX,MY,NX,NY))
    sys.exit(2)
  # return
  r_arc_center_radius=(IX, IY, IA)
  return(r_arc_center_radius)

def arc_center_radius_angles(ai_arc_pt1, ai_arc_pt2, ai_arc_pt3, ai_error_msg_id):
  """ Compute the center, radius and angles of the arc defined by the three points A,B and C
  """
  # use to check is angle is smaller than pi/2
  radian_epsilon = math.pi/1000
  # interpretation of the input points
  AX = ai_arc_pt1[0]
  AY = ai_arc_pt1[1]
  BX = ai_arc_pt2[0]
  BY = ai_arc_pt2[1]
  CX = ai_arc_pt3[0]
  CY = ai_arc_pt3[1]
  # calculation of I
  (IX,IY, arc_radius) = arc_center_radius(ai_arc_pt1, ai_arc_pt2, ai_arc_pt3, ai_error_msg_id)
  # check I is equidistant of A,B,C,D,E
  IA = math.sqrt((AX-IX)**2+(AY-IY)**2)
  IB = math.sqrt((BX-IX)**2+(BY-IY)**2)
  IC = math.sqrt((CX-IX)**2+(CY-IY)**2)
  if((abs(IA-arc_radius)>radian_epsilon)or(abs(IB-arc_radius)>radian_epsilon)or(abs(IC-arc_radius)>radian_epsilon)):
    print("ERR841: Error, in {:s}, I is not equidistant from A,B,C. arc_radius={:0.2f} IA={:0.2f} IB={:0.2f} IC={:0.2f}".format(ai_error_msg_id, arc_radius, IA, IB, IC))
    sys.exit(2)
  # calculation of the angle u=(Ix, IA) , v=(Ix, IB), w=(Ix, IC), d=(Ix, ID) and e=(Ix, IE)
  u = math.atan2(AY-IY, AX-IX)
  v = math.atan2(BY-IY, BX-IX)
  w = math.atan2(CY-IY, CX-IX)
  # calculation of the angle uv=(IA, IB), uw=(IA, IC)
  uv = math.fmod(v-u+4*math.pi, 2*math.pi)
  uw = math.fmod(w-u+4*math.pi, 2*math.pi)
  # check arc direction
  ccw_ncw = True
  if(uw>uv):
    #print("dbg874: arc of circle direction: counter clock wise (CCW)")
    ccw_ncw = True
  else:
    #print("dbg875: arc of circle direction: clock wise (CW)")
    ccw_ncw = False
    uv = uv - 2*math.pi
    uw = uw - 2*math.pi
  # return
  r_arc_center_radius_angles=(IX, IY, IA, uw, u, w)
  return(r_arc_center_radius_angles)

# aka circle_circle_intersection
def triangulation(ai_A, ai_AC, ai_B, ai_BC, ai_D, ai_D_direction, ai_error_msg_id):
  """ knowing the coordiantes of A and B and the lengths AC and BC, returns the coordinates of C
      C is placed on the same side as D compare to the line (AB)
      If D is too closed to the line (AB), the D_direction is used to identified the side to place C
  """
  # use to check is angle is smaller than pi/2
  radian_epsilon = math.pi/1000
  # code error
  r_status = 0
  # check the arguments
  if((ai_AC<radian_epsilon)or(ai_BC<radian_epsilon)):
    print("ERR965: Error in {:s}, the length ai_AC (={:0.2f}) or ai_BC (={:0.2f})".format(ai_error_msg_id, ai_AC, ai_BC))
    sys.exir(2)
  # interprete the arguments
  AX = ai_A[0]
  AY = ai_A[1]
  b = ai_AC
  BX = ai_B[0]
  BY = ai_B[1]
  a = ai_BC
  DX = ai_D[0]
  DY = ai_D[1]
  # calculation of the length c=AB
  c = math.sqrt((BX-AX)**2+(BY-AY)**2)
  if(c<radian_epsilon):
    print("ERR662: Error in {:s}, the length c (=AB={:0.2f}) is too small!".format(ai_error_msg_id, c))
    sys.exit(2)
  # calculation of the angle A with the law of cosines
  #BAC = math.acos((b**2+c**2-a**2)/(2*b*c))
  cos_BAC = (b**2+c**2-a**2)/(2*b*c)
  if(abs(cos_BAC)>1):
    print("ERR542: Error in {:s}, cos_BAC is out of [-1,1]! a={:0.2f} b={:0.2f} c={:0.2f} cos_BAC={:0.2f}".format(ai_error_msg_id, a, b, c, cos_BAC))
    print("dbg652: AX={:0.2f} AY={:0.2f} BX={:0.2f} BY={:0.2f}".format(AX,AY,BX,BY))
    #sys.exit(2)
    r_status=2
    return((0,0,r_status))
  BAC = math.acos(cos_BAC)
  # calculation of the angle xAB
  xAB = math.atan2(BY-AY, BX-AX)
  # calculation of the angle BAD
  xAD =  math.atan2(DY-AY, DX-AX) # atan2 avoid the problem of divided by zero
  BAD = math.fmod(xAD - xAB + 5*math.pi, 2*math.pi) - math.pi
  if(abs(BAD)<radian_epsilon):
    print("WARN546: Warning in {:s}, the side of the triangulation is not clear! Use the ai_D_direction".format(ai_error_msg_id))
    DX1 = DX+5*radian_epsilon*math.cos(ai_D_direction)
    DY1 = DY+5*radian_epsilon*math.sin(ai_D_direction)
    xAD =  math.atan2(DY1-AY, DX1-AX) # atan2 avoid the problem of divided by zero
    BAD = math.fmod(xAD - xAB + 5*math.pi, 2*math.pi) - math.pi
  # calculation of the angle xAC
  xAC = xAB + math.copysign(BAC, BAD)
  # calculation of the coordinates of C
  CX = AX+b*math.cos(xAC)
  CY = AY+b*math.sin(xAC)
  # for verification, duplication of the calculation via B
  ABC = math.acos((a**2+c**2-b**2)/(2*a*c))
  xBA = math.atan2(AY-BY, AX-BX) # = xAB+math.pi
  #BD = math.sqrt((DX-BX)**2+(DY-BY)**2)
  xBD = math.atan2(DY-BY, DX-BX)
  ABD = math.fmod(xBD - xBA + 5*math.pi, 2*math.pi) - math.pi
  if(abs(ABD)<radian_epsilon):
    print("WARN547: Warning in {:s}, the side of the triangulation is not clear! Use the ai_D_direction".format(ai_error_msg_id))
    DX1 = DX+5*radian_epsilon*math.cos(ai_D_direction)
    DY1 = DY+5*radian_epsilon*math.sin(ai_D_direction)
    xBD = math.atan2(DY1-BY, DX1-BX)
    ABD = math.fmod(xBD - xBA + 5*math.pi, 2*math.pi) - math.pi
  xBC = xBA + math.copysign(ABC, ABD)
  CX2 = BX+a*math.cos(xBC)
  CY2 = BY+a*math.sin(xBC)
  if((abs(CX2-CX)>radian_epsilon)or(abs(CY2-CY)>radian_epsilon)):
    print("ERR686: Error in {:s}, the coordinate of C seems wrong! CX={:0.2f} CX2={:0.2f} CY={:0.2f} CY2={:0.2f}".format(ai_error_msg_id, CX, CX2, CY, CY2))
    print("dbg545: D {:0.2f} {:0.2f}  ai_D_direction {:0.2f}".format(DX,DY, ai_D_direction))
    print("dbg512: BAC", BAC)
    print("dbg513: xAB", xAB)
    #print("dbg514: AD", AD)
    print("dbg515: BAD", BAD)
    print("dbg516: xAC", xAC)
    print("dbg517: AX {:0.2f}  AY {:0.2f}  b {:0.2f}".format(AX,AY, b))
    print("dbg522: ABC", ABC)
    print("dbg523: xBA", xBA)
    #print("dbg524: BD", BD)
    print("dbg525: ABD", ABD)
    print("dbg526: xBC", xBC)
    print("dbg527: BX {:0.2f}  BY {:0.2f}  a {:0.2f}".format(BX,BY, a))
    sys.exit(2)
  r_C = (CX,CY,r_status)
  return(r_C)

def line_equation(ai_A, ai_B, ai_error_msg_id):
  """ Given the coordinates of two points, it returns the three coefficient of the line equation, the length of the segment and the inclination
  """
  # use to check is angle is smaller than pi/2
  radian_epsilon = math.pi/1000
  # interpretation of the input points
  AX = ai_A[0]
  AY = ai_A[1]
  BX = ai_B[0]
  BY = ai_B[1]
  # calculation of the length AB
  lAB = math.sqrt((BX-AX)**2+(BY-AY)**2)
  # calculation of the inclination of AB
  xAB = math.atan2(BY-AY, BX-AX)
  cos_xAB = (BX-AX)/lAB
  sin_xAB = (BY-AY)/lAB
  # calculation of the line equation coefficient : ABlx * x + ABly * y + ABk = 0
  # ABlx = sin(xAB)
  ABlx = sin_xAB
  # ABly = -cos(xAB)
  ABly = -cos_xAB
  # ABk = -(ABlx*AX+ABly*AY)
  ABk = -(ABlx*AX+ABly*AY)
  # return
  r_line_equation = (ABlx, ABly, ABk, lAB, xAB)
  return(r_line_equation)

def line_distance_point(ai_A, ai_B, ai_q, ai_error_msg_id):
  """ Given the two points A and B and the distance q, what are the coordinates of Q, distance of q from A and AB
  """
  # use to check is angle is smaller than pi/2
  radian_epsilon = math.pi/1000
  # interpretation of the input points
  AX = ai_A[0]
  AY = ai_A[1]
  BX = ai_B[0]
  BY = ai_B[1]
  # line equation of AB
  #(ABlx, ABly, ABkA, lAB, xAB) = line_equation(ai_A, ai_B, ai_error_msg_id)
  lAB = math.sqrt((BX-AX)**2+(BY-AY)**2)
  cos_xAB = (BX-AX)/lAB
  sin_xAB = (BY-AY)/lAB
  ABlx = sin_xAB
  ABly = -cos_xAB
  # calculation of Q
  # QX = AX+q*cos(xAB+pi/2) = AX-q*sin(xAB)
  # QY = AX+q*sin(xAB+pi/2) = AX+q*cos(xAB)
  QX = AX-ai_q*sin_xAB
  QY = AY+ai_q*cos_xAB
  ABkQ = -(ABlx*QX+ABly*QY)
  # return
  r_Q = (QX, QY, ABkQ)
  return(r_Q)

def line_point_projection(ai_AB, ai_M, ai_error_msg_id):
  """ Given the line equation of AB (ABlx, AVly, ABk) and the coordinates of M (MX,MY), returns the coordinates of P, projection of M on AB
  """
  # use to check is angle is smaller than pi/2
  #radian_epsilon = math.pi/1000
  # interpretation of the input points
  a = ai_AB[0] # ABlx
  b = ai_AB[1] # ABly
  c = ai_AB[2] # ABk
  MX = ai_M[0]
  MY = ai_M[1]
  # calculation of d (MPk, coefficient of the line equation of PM)
  d = -(b*MX-a*MY) # MPk
  # calculation of the coordinates of P
  PX = -(c*a+d*b)
  PY = d*a-c*b
  # return
  r_P = (PX, PY)
  return(r_P)

def line_circle_intersection(ai_AB, ai_I, ai_R, ai_C, ai_D_direction, ai_error_msg_id):
  """ Given the line equation (a*x+b*y+c=0) and the circle of center I and radius R, returns the intersection M
      C define the side of the intersection
      D_direction is used if C is too closed to the border. Set it to zero if you don't know how to set it.
  """
  # use to check is angle is smaller than pi/2
  radian_epsilon = math.pi/1000
  # status for error forward
  line_circle_intersection_status = 0
  r_line_circle_intersection = (0, 0, line_circle_intersection_status)
  # interpretation of the input points
  a = ai_AB[0] # ABlx
  b = ai_AB[1] # ABly
  c = ai_AB[2] # ABk
  IX = ai_I[0]
  IY = ai_I[1]
  CX = ai_C[0]
  CY = ai_C[1]
  # calculation of the coordinates of P, projection of I on AB
  (PX, PY) = line_point_projection(ai_AB, ai_I, ai_error_msg_id)
  # calculation of the coordinates of C2, projection of C on AB
  (C2X, C2Y) = line_point_projection(ai_AB, ai_C, ai_error_msg_id)
  # calculation of the length IP
  #IP = math.sqrt((PX-IX)**2+(PY-IY)**2)
  IP2 = (PX-IX)**2+(PY-IY)**2
  IP = math.sqrt(IP2)
  if(IP>ai_R):
    # This error might occur when the sign of the tangent is wrong because of calculation impression.
    # That's why, we need to report the error instead of stopping on this error
    print("ERR672: Error in {:s}, the line and the circle have no intersection! IP={:0.2f}  ai_R={:0.2f}".format(ai_error_msg_id, IP, ai_R))
    print("dbg611: ai_AB:", ai_AB)
    print("dbg612: ai_I:", ai_I)
    print("dbg613: ai_R:", ai_R)
    print("dbg614: ai_C:", ai_C)
    #sys.exit(2)
    line_circle_intersection_status=2
    r_line_circle_intersection = (0, 0, line_circle_intersection_status)
  else:
    # calculation of the length PM
    PM = math.sqrt(ai_R**2-IP2)
    # calculation of the length C2P
    C2P = math.sqrt((PX-C2X)**2+(PY-C2Y)**2)
    cos_i = (C2X-PX)/C2P
    sin_i = (C2Y-PY)/C2P
    if(C2P<radian_epsilon):
      print("WARN348: Warning in {:s}, C is too closed from the board to find the line_circle_intersection!".format(ai_error_msg_id))
      xAB = math.atan2(a, -b)
      direction_correlation = math.fmod(ai_D_direction-xAB+5*math.pi, 2*math.pi)-math.pi
      direction_AB = math.copysign(1, direction_correlation)
      norm_ab = math.sqrt(a**2+b**2)
      cos_i = direction_AB * -b / norm_ab
      sin_i = direction_AB * a / norm_ab
    # calculation of the coordinates of M
    MX = PX+cos_i*PM
    MY = PY+sin_i*PM
    r_line_circle_intersection = (MX, MY, line_circle_intersection_status)
  # return
  return(r_line_circle_intersection)

def line_line_intersection(ai_AB, ai_CD, ai_error_msg_id):
  """ Given the line equations (a*x+b*y+c=0) of the two lines (AB) and (CD), returns the intersection M
  """
  # use to check is angle is smaller than pi/2
  radian_epsilon = math.pi/1000
  # status for error forward
  line_line_intersection_status = 0
  # interpretation of the input points
  a1 = ai_AB[0] # ABlx
  b1 = ai_AB[1] # ABly
  c1 = ai_AB[2] # ABk
  a2 = ai_CD[0] # CDlx
  b2 = ai_CD[1] # CDly
  c2 = ai_CD[2] # CDk
  # we have
  # a1*MX+b1*MY+c1=0
  # a2*MX+b2*MY+c2=0
  MX = 0
  MY = 0
  determinant = a1*b2-a2*b1
  if(abs(determinant)<radian_epsilon):
    print("WARN656: Warning in {:s}, the two lines are parallel!".format(ai_error_msg_id))
    line_line_intersection_status=2
  else:
    MX = (c2*b1-c1*b2)/determinant
    MY = (c1*a2-c2*a1)/determinant
  r_line_line_intersection = (MX, MY, line_line_intersection_status)
  # return
  #print("dbg947: r_line_line_intersection:", r_line_line_intersection)
  return(r_line_line_intersection)

def sub_smooth_corner_line_arc(ai_pre_point, ai_current_point, ai_post_middle, ai_post_point, ai_router_bit_request, ai_error_msg_id):
  """ compute the corner center of a smoothed line-arc corner
  """
  # use to check is angle is smaller than pi/2
  radian_epsilon = math.pi/1000
  # interpretation of the input points
  #AX = ai_pre_point[0]
  #AY = ai_pre_point[1]
  CX = ai_current_point[0]
  CY = ai_current_point[1]
  #DX = ai_post_middle[0]
  #DY = ai_post_middle[1]
  #EX = ai_post_point[0]
  #EY = ai_post_point[1]
  # calculation of the AC line equation and J
  (AClx, ACly, ACkA, lAC, xAC) = line_equation(ai_pre_point, ai_current_point, ai_error_msg_id)
  (JX,JY, R2, uw2, u2, w2) = arc_center_radius_angles(ai_current_point, ai_post_middle, ai_post_point, ai_error_msg_id)
  # arc orientation
  #o1 = math.copysign(1, uw1)
  o2 = math.copysign(1, uw2)
  # sign of the tangent angle
  tangent_angle = math.fmod( u2+o2*math.pi/2-xAC+7*math.pi, 2*math.pi) - math.pi
  #r_outline = []
  smooth_status = 0
  if(abs(tangent_angle)<radian_epsilon):
    print("WARN942: Warning in {:s}, the tangent_angle is too flat! the corner doesn't need to be smoothed.".format(ai_error_msg_id))
    #r_outline = [(ai_current_point[0], ai_current_point[1])]
    smooth_status = 2
  #elif(abs(tangent_angle)>math.pi-radian_epsilon):
  #  print("WARN943: Warning in {:s}, the tangent_angle is too sharp! the corner cannot be smoothed.".format(ai_error_msg_id))
  #  r_outline = [(ai_current_point[0], ai_current_point[1])]
  else:
    o3 = math.copysign(1, tangent_angle)
    # calculation of IS and JS
    Q1_plus = o3 # = -1 if the router below the line, else 1
    R2_plus = o2*o3 # = -1 if the router is outer the arc, else 1
    (QX,QY, ACkQ) = line_distance_point(ai_pre_point, ai_current_point, Q1_plus*ai_router_bit_request, ai_error_msg_id)
    JS = R2-R2_plus*ai_router_bit_request
    #print("dbg147: ai_pre_point", ai_pre_point)
    #print("dbg347: ai_current_point", ai_current_point)
    #print("dbg447: ai_post_middle", ai_post_middle)
    #print("dbg547: ai_post_point", ai_post_point)
    #print("dbg647: ai_router_bit_request", ai_router_bit_request)
    #print("dbg123: AClx={:0.2f} ACly={:0.2f} ACkA={:0.2f} lAC={:0.2f} xAC={:0.2f}".format(AClx, ACly, ACkA, lAC, xAC))
    #print("dbg124: JX={:0.2f} JY={:0.2f} R2={:0.2f} uw2={:0.2f} u2={:0.2f}  w2={:0.2f}".format(JX, JY, R2, uw2, u2, w2))
    #print("dbg125: o2={:0.2f} o3={:0.2f}".format(o2,o3))
    #print("dbg127: tangent_angle", tangent_angle)
    #print("dbg126: QS={:0.2f} JS={:0.2f}".format(QS,JS))
    #print("dbg532: in {:s}, xAC={:0.2f} R2={:0.2f} ai_router_bit_request={:0.2f}".format(ai_error_msg_id, xAC, R2, ai_router_bit_request))
    # calculation of the coordiantes of S, the center of the router_bit in the smooth corner
    bisector_angle = math.fmod(u2+o2*math.pi/2-(xAC+math.pi) + 7*math.pi, 2*math.pi) - math.pi
    D_direction = xAC + bisector_angle/2
    #print("dbg693: bisector_angle {:0.2f}  D_direction {:0.2f}".format(bisector_angle, D_direction))
    (SX,SY, line_circle_intersection_status) = line_circle_intersection((AClx, ACly, ACkQ), (JX,JY),JS, (CX,CY), D_direction, ai_error_msg_id)
    # when the angle is too sharp, the sign of the tangeant angle might be wrong because of calculation imprecition.
    # in this case, we use the method of try and retry
    if((line_circle_intersection_status==2)and(abs(tangent_angle)>math.pi-10*radian_epsilon)): # error then retry with -Q1_plus instead of Q1_plus
      print("WARN682: Warning in {:s}, line_arc corner is smoothed with the other side because of a line_circle_intersection error!".format(ai_error_msg_id))
      (QX,QY, ACkQ) = line_distance_point(ai_pre_point, ai_current_point, -Q1_plus*ai_router_bit_request, ai_error_msg_id)
      tmpCX = CX + JS*math.cos(xAC+math.pi)
      tmpCY = CY + JS*math.sin(xAC+math.pi)
      (SX,SY, line_circle_intersection_status) = line_circle_intersection((AClx, ACly, ACkQ), (JX,JY),JS, (tmpCX,tmpCY), D_direction, ai_error_msg_id)
    # end of the retry. Continue the normal calculation recipe
    if(line_circle_intersection_status==2):
      print("WARN684: Warning in {:s}, corner is not smoothed because of a line_circle_intersection error!".format(ai_error_msg_id))
      print("dbg681: ai_pre_point", ai_pre_point)
      print("dbg683: ai_current_point", ai_current_point)
      print("dbg684: ai_post_middle", ai_post_middle)
      print("dbg685: ai_post_point", ai_post_point)
      print("dbg686: ai_router_bit_request", ai_router_bit_request)
      print("dbg623: AClx={:0.2f} ACly={:0.2f} ACkA={:0.2f} lAC={:0.2f} xAC={:0.2f}".format(AClx, ACly, ACkA, lAC, xAC))
      print("dbg688: JX={:0.2f} JY={:0.2f} R2={:0.2f} uw2={:0.2f} u2={:0.2f}  w2={:0.2f}".format(JX, JY, R2, uw2, u2, w2))
      print("dbg689: o2={:0.2f} o3={:0.2f}".format(o2,o3))
      print("dbg691: tangent_angle", tangent_angle)
      print("dbg692: JS={:0.2f}".format(JS))
      print("dbg711: QX={:0.2f} QY={:0.2f} ACkQ={:0.2f}".format(QX,QY,ACkQ))
      print("dbg693: in {:s}, xAC={:0.2f} R2={:0.2f} ai_router_bit_request={:0.2f}".format(ai_error_msg_id, xAC, R2, ai_router_bit_request))
      #r_outline = [(ai_current_point[0], ai_current_point[1])]
      smooth_status = 2
      sys.exit(2)
    else:
      # calculation of U, the projection of S on AC
      (UX, UY) = line_point_projection((AClx, ACly, ACkA), (SX,SY), ai_error_msg_id)
      # calculation of the angles xSU and xSJ
      #xSU = math.atan2(UY-SY, UX-SX)+(1+Q1_plus)/2*math.pi
      xSU = math.atan2(UY-SY, UX-SX)
      xSJ = math.atan2(JY-SY, JX-SX)+(1+R2_plus)/2*math.pi
      router_bit_arc_uw = math.fmod(xSJ-xSU+4*math.pi, 2*math.pi)
      if(o3<0):
        router_bit_arc_uw = router_bit_arc_uw - 2*math.pi
      #print("dbg337: SX {:0.2f}  SY {:0.2f}".format(SX,SY))
      #print("dbg994: xSI {:0.2f}  xSJ {:0.2f}".format(xSI, xSJ))
      #print("dbg773: router_bit_arc_uw", router_bit_arc_uw)
      smooth_status = 1
  # return
  r_sub_smooth_corner_line_arc = (SX, SY, UX, UY, xSU, xSJ, router_bit_arc_uw, smooth_status)
  return(r_sub_smooth_corner_line_arc)

def sub_smooth_corner_arc_arc(ai_pre_point, ai_pre_middle, ai_current_point, ai_post_middle, ai_post_point, ai_router_bit_request, ai_error_msg_id):
  """ Compute the smooth corner center for a arc-arc corner
  """
  # use to check is angle is smaller than pi/2
  radian_epsilon = math.pi/1000
  # interpretation of the input points
  #AX = ai_pre_point[0]
  #AY = ai_pre_point[1]
  #BX = ai_pre_middle[0]
  #BY = ai_pre_middle[1]
  CX = ai_current_point[0]
  CY = ai_current_point[1]
  #DX = ai_post_middle[0]
  #DY = ai_post_middle[1]
  #EX = ai_post_point[0]
  #EY = ai_post_point[1]
  # calculation of I and J
  (IX,IY, R1, uw1, u1, w1) = arc_center_radius_angles(ai_pre_point, ai_pre_middle, ai_current_point, ai_error_msg_id)
  (JX,JY, R2, uw2, u2, w2) = arc_center_radius_angles(ai_current_point, ai_post_middle, ai_post_point, ai_error_msg_id)
  # arc orientation
  o1 = math.copysign(1, uw1)
  o2 = math.copysign(1, uw2)
  # sign of the tangent angle
  #tangent_angle = math.fmod( (u2+o2*path.pi/2)-(w1+o1*path.pi/2)+8*math.pi, 2*math.pi) - math.pi
  tangent_angle = math.fmod( u2-w1+(o2-o1)*math.pi/2+9*math.pi, 2*math.pi) - math.pi
  #r_outline = []
  smooth_status = 0
  if(abs(tangent_angle)<radian_epsilon):
    print("WARN932: Warning in {:s}, the tangent_angle is too flat! the corner doesn't need to be smoothed.".format(ai_error_msg_id))
    #r_outline = [(ai_current_point[0], ai_current_point[1])]
    smooth_status = 2
  elif(abs(tangent_angle)>math.pi-radian_epsilon):
    print("WARN933: Warning in {:s}, the tangent_angle is too sharp! the corner cannot be smoothed.".format(ai_error_msg_id))
    #r_outline = [(ai_current_point[0], ai_current_point[1])]
    smooth_status = 2
  else:
    o3 = math.copysign(1, tangent_angle)
    # calculation of IS and JS
    R1_plus = o1*o3 # = -1 if the router is outer the arc, else 1
    R2_plus = o2*o3 # = -1 if the router is outer the arc, else 1
    IS = R1-R1_plus*ai_router_bit_request
    JS = R2-R2_plus*ai_router_bit_request
    #print("dbg147: ai_pre_point", ai_pre_point)
    #print("dbg247: ai_pre_middle", ai_pre_middle)
    #print("dbg347: ai_current_point", ai_current_point)
    #print("dbg447: ai_post_middle", ai_post_middle)
    #print("dbg547: ai_post_point", ai_post_point)
    #print("dbg647: ai_router_bit_request", ai_router_bit_request)
    #print("dbg123: IX={:0.2f} IY={:0.2f} R1={:0.2f} uw1={:0.2f} u1={:0.2f}  w1={:0.2f}".format(IX, IY, R1, uw1, u1, w1))
    #print("dbg124: JX={:0.2f} JY={:0.2f} R2={:0.2f} uw2={:0.2f} u2={:0.2f}  w2={:0.2f}".format(JX, JY, R2, uw2, u2, w2))
    #print("dbg125: o1={:0.2f} o2={:0.2f} o3={:0.2f}".format(o1,o2,o3))
    #print("dbg127: tangent_angle", tangent_angle)
    #print("dbg126: IS={:0.2f} JS={:0.2f}".format(IS,JS))
    #print("dbg532: in {:s}, R1={:0.2f} R2={:0.2f} ai_router_bit_request={:0.2f}".format(ai_error_msg_id, R1, R2, ai_router_bit_request))
    # calculation of the coordiantes of S, the center of the router_bit in the smooth corner
    bisector_angle = math.fmod(u2-w1+(o2+o1)*math.pi/2 + 9*math.pi, 2*math.pi) - math.pi #(u2+o2*math.pi/2)-(w1-o1*math.pi/2)
    D_direction = w1 + bisector_angle/2
    #print("dbg693: bisector_angle {:0.2f}  D_direction {:0.2f}".format(bisector_angle, D_direction))
    (SX,SY, triangulation_status) = triangulation((IX,IY),IS,(JX,JY),JS,(CX,CY), D_direction, ai_error_msg_id)
    if(triangulation_status==2):
      print("WARN693: Warning in {:s}, corner is not smoothed because of a triangulation error!".format(ai_error_msg_id))
      #print("dbg681: ai_pre_point", ai_pre_point)
      #print("dbg682: ai_pre_middle", ai_pre_middle)
      #print("dbg683: ai_current_point", ai_current_point)
      #print("dbg684: ai_post_middle", ai_post_middle)
      #print("dbg685: ai_post_point", ai_post_point)
      #print("dbg686: ai_router_bit_request", ai_router_bit_request)
      #print("dbg687: IX={:0.2f} IY={:0.2f} R1={:0.2f} uw1={:0.2f} u1={:0.2f}  w1={:0.2f}".format(IX, IY, R1, uw1, u1, w1))
      #print("dbg688: JX={:0.2f} JY={:0.2f} R2={:0.2f} uw2={:0.2f} u2={:0.2f}  w2={:0.2f}".format(JX, JY, R2, uw2, u2, w2))
      #print("dbg689: o1={:0.2f} o2={:0.2f} o3={:0.2f}".format(o1,o2,o3))
      #print("dbg691: tangent_angle", tangent_angle)
      #print("dbg692: IS={:0.2f} JS={:0.2f}".format(IS,JS))
      #print("dbg693: in {:s}, R1={:0.2f} R2={:0.2f} ai_router_bit_request={:0.2f}".format(ai_error_msg_id, R1, R2, ai_router_bit_request))
      #r_outline = [(ai_current_point[0], ai_current_point[1])]
      smooth_status = 2
    else:
      # calculation of the angles xSI and xSJ
      xSI = math.atan2(IY-SY, IX-SX)+(1+R1_plus)/2*math.pi
      xSJ = math.atan2(JY-SY, JX-SX)+(1+R2_plus)/2*math.pi
      router_bit_arc_uw = math.fmod(xSJ-xSI+4*math.pi, 2*math.pi)
      if(o3<0):
        router_bit_arc_uw = router_bit_arc_uw - 2*math.pi
      #print("dbg337: SX {:0.2f}  SY {:0.2f}".format(SX,SY))
      #print("dbg994: xSI {:0.2f}  xSJ {:0.2f}".format(xSI, xSJ))
      #print("dbg773: router_bit_arc_uw", router_bit_arc_uw)
      smooth_status = 1
  #return
  r_sub_smooth_corner_arc_arc = (SX, SY, xSI, xSJ, router_bit_arc_uw, smooth_status)
  return(r_sub_smooth_corner_arc_arc)

def sub_enlarge_corner_line_arc(ai_pre_point, ai_current_point, ai_post_middle, ai_post_point, ai_router_bit_request, ai_error_msg_id):
  """ Compute the points related to an enlarged line-arc corner
  """
  # use to check is angle is smaller than pi/2
  radian_epsilon = math.pi/1000
  # interpretation of the input points
  #AX = ai_pre_point[0]
  #AY = ai_pre_point[1]
  CX = ai_current_point[0]
  CY = ai_current_point[1]
  #DX = ai_post_middle[0]
  #DY = ai_post_middle[1]
  #EX = ai_post_point[0]
  #EY = ai_post_point[1]
  enlarge_status = 0
  #r_sub_enlarge_corner_line_arc = ((MKX,MKY), (GX,GY), (CX,CY), (HX,HY), (NLX,NLY), enlarge_type_request_1, enlarge_type_request_2, enlarge_status)
  MKX=0; MKY=0; GX=0; GY=0; HX=0; HY=0; NLX=0; NLY=0; enlarge_type_request_1=0; enlarge_type_request_2=0
  (SX, SY, UX, UY, xSU, xSJ, router_bit_arc_uw, smooth_status) = sub_smooth_corner_line_arc(ai_pre_point, ai_current_point, ai_post_middle, ai_post_point, ai_router_bit_request, ai_error_msg_id)
  if(smooth_status!=1):
    print("ERR977: Error in {:s}, the sub smooth corner get an error!".format(ai_error_msg_id))
    sys.exit(2)
    enlarge_status = 2
  else:
    #print("dbg553: router_bit_arc_uw:", router_bit_arc_uw)
    if(abs(router_bit_arc_uw)>math.pi):
      print("ERR877: Error in {:s}, the sub smooth corner is englobed!".format(ai_error_msg_id))
      sys.exit(2)
      enlarge_status = 2
    elif(abs(router_bit_arc_uw)<radian_epsilon):
      print("WARN878: Warning in {:s}, the corner is too flat to be enlarged with this router_bit radius!")
      enlarge_status = 2
      enlarge_type_request_1=1
      enlarge_type_request_2=1
    else:
      corner_orientation = math.copysign(1, router_bit_arc_uw)
      enlarge_status = 1
      (SClx, SCly, SCkS, lSC, xSC) = line_equation((SX,SY), (CX,CY), ai_error_msg_id)
      (FX,FY, line_circle_intersection_status) = line_circle_intersection((SClx, SCly, SCkS), (CX,CY),ai_router_bit_request, (SX,SY), xSC+math.pi, ai_error_msg_id)
      if(line_circle_intersection_status==2):
        print("ERR739: Error in {:s} with the line_circle_intersection_status!".format(ai_error_msg_id))
        sys.exit(2)
      (GX,GY, SCkG) = line_distance_point((FX,FY), (CX,CY), -1*corner_orientation*ai_router_bit_request, ai_error_msg_id)
      (HX,HY, SCkH) = line_distance_point((FX,FY), (CX,CY),  1*corner_orientation*ai_router_bit_request, ai_error_msg_id)
      #(IX,IY, R1, uw1, u1, w1) = arc_center_radius_angles(ai_pre_point, ai_pre_middle, ai_current_point, ai_error_msg_id)
      (AClx, ACly, ACkC, lAC, xAC) = line_equation(ai_pre_point, ai_current_point, ai_error_msg_id)

      (JX,JY, R2, uw2, u2, w2) = arc_center_radius_angles(ai_current_point, ai_post_middle, ai_post_point, ai_error_msg_id)
      #(MX,MY, line_circle_intersection_status) = line_circle_intersection((SClx, SCly, SCkG), (IX,IY),R1, ((GX+SX)/2,(GY+SY)/2), xSC+math.pi, ai_error_msg_id)
      (MX,MY, line_line_intersection_status) = line_line_intersection((SClx, SCly, SCkG), (AClx, ACly, ACkC), ai_error_msg_id)

      #print("dbg321: MX={:0.2f}  MY={:0.2f}  line_circle_intersection={:0.2f}".format(MX, MY, line_circle_intersection_status))
      if(line_line_intersection_status==2):
        print("ERR736: Error in {:s} with the line_line_intersection_status!".format(ai_error_msg_id))
        sys.exit(2)
        enlarge_status = 2
      (NX,NY, line_circle_intersection_status) = line_circle_intersection((SClx, SCly, SCkH), (JX,JY),R2, ((HX+SX)/2,(HY+SY)/2), xSC+math.pi, ai_error_msg_id)
      #print("dbg322: NX={:0.2f}  NY={:0.2f}  line_circle_intersection={:0.2f}".format(NX, NY, line_circle_intersection_status))
      if(line_circle_intersection_status==2):
        print("ERR733: Error in {:s} with the line_circle_intersection_status!".format(ai_error_msg_id))
        #print("dbg513: G={:0.2f} {:0.2f}  H={:0.2f} {:0.2f}".format(GX, GY, HX, HY))
        sys.exit(2)
        enlarge_status = 2
      # check if arc-arc intersection must be calculated
      tmp_MG_deep_x = (MX-GX)*(CX-FX)
      tmp_MG_deep_y = (MY-GY)*(CY-FY)
      #print("dbg301: tmp_MG_deep_x:", tmp_MG_deep_x)
      #print("dbg302: tmp_MG_deep_y:", tmp_MG_deep_y)
      MG_deep = math.copysign(1, tmp_MG_deep_x)
      if(abs(tmp_MG_deep_x)<abs(tmp_MG_deep_y)):
        MG_deep = math.copysign(1, tmp_MG_deep_y)
        #print("dbg303: use Y to get MG_deep")
      tmp_NH_deep_x = (NX-HX)*(CX-FX)
      tmp_NH_deep_y = (NY-HY)*(CY-FY)
      #print("dbg311: tmp_NH_deep_x:", tmp_NH_deep_x)
      #print("dbg312: tmp_NH_deep_y:", tmp_NH_deep_y)
      NH_deep = math.copysign(1, tmp_NH_deep_x)
      if(abs(tmp_NH_deep_x)<abs(tmp_NH_deep_y)):
        NH_deep = math.copysign(1, tmp_NH_deep_y)
        #print("dbg313: use Y to get NH_deep")
      MKX = MX
      MKY = MY
      NLX = NX
      NLY = NY
      enlarge_type_request_1 = 3
      enlarge_type_request_2 = 3
      # compute the arc-arc intersection
      if(MG_deep>0):
        #(KX,KY, triangulation_status) = triangulation((IX,IY),R1,(FX,FY),ai_router_bit_request,(GX,GY), xSC+math.pi, ai_error_msg_id)
        (KX,KY, line_circle_intersection_status) = line_circle_intersection((AClx, ACly, ACkC), (FX,FY),ai_router_bit_request, (GX, GY), xSC+math.pi, ai_error_msg_id)

        if(line_circle_intersection==2):
          print("ERR536: Error in {:s} with the triangulation_status!".format(ai_error_msg_id))
          sys.exit(2)
        MKX = KX
        MKY = KY
        enlarge_type_request_1 = 2
      if(NH_deep>0):
        (LX,LY, triangulation_status) = triangulation((JX,JY),R2, (FX,FY),ai_router_bit_request, (HX,HY), xSC+math.pi, ai_error_msg_id)
        if(triangulation_status==2):
          print("ERR533: Error in {:s} with the triangulation_status!".format(ai_error_msg_id))
          sys.exit(2)
        NLX = LX
        NLY = LY
        enlarge_type_request_2 = 2
      lMN=math.sqrt((NLX-MKX)**2+(NLY-MKY)**2)
      if(lMN<radian_epsilon):
        print("WARN667: Warning in {:s}, the angle is too flat to be enlarged!".format(ai_error_msg_id))
        enlarge_type_request_1 = 1
        enlarge_type_request_2 = 1
  # return
  r_sub_enlarge_corner_line_arc = ((MKX,MKY), (GX,GY), (CX,CY), (HX,HY), (NLX,NLY), enlarge_type_request_1, enlarge_type_request_2, enlarge_status)
  #print("dbg668: r_sub_enlarge_corner_line_arc:", r_sub_enlarge_corner_line_arc)
  return(r_sub_enlarge_corner_line_arc)

def sub_enlarge_corner_arc_arc(ai_pre_point, ai_pre_middle, ai_current_point, ai_post_middle, ai_post_point, ai_router_bit_request, ai_error_msg_id):
  """ Compute the points related to an enlarged arc-arc corner
  """
  # use to check is angle is smaller than pi/2
  radian_epsilon = math.pi/1000
  # interpretation of the input points
  #AX = ai_pre_point[0]
  #AY = ai_pre_point[1]
  #BX = ai_pre_middle[0]
  #BY = ai_pre_middle[1]
  CX = ai_current_point[0]
  CY = ai_current_point[1]
  #DX = ai_post_middle[0]
  #DY = ai_post_middle[1]
  #EX = ai_post_point[0]
  #EY = ai_post_point[1]
  #r_sub_enlarge_corner_arc_arc = ((MKX,MKY), (GX,GY), (CX,CY), (HX,HY), (NLX,NLY), enlarge_type_request_1, enlarge_type_request_2, enlarge_status)
  MKX=0; MKY=0; GX=0; GY=0; HX=0; HY=0; NLX=0; NLY=0; enlarge_type_request_1=0; enlarge_type_request_2=0
  (SX, SY, xSI, xSJ, router_bit_arc_uw, smooth_status) = sub_smooth_corner_arc_arc(ai_pre_point, ai_pre_middle, ai_current_point, ai_post_middle, ai_post_point, ai_router_bit_request, ai_error_msg_id)
  enlarge_status = 0
  if(smooth_status==2):
    print("ERR927: Error in {:s}, the sub smooth corner gets an error!".format(ai_error_msg_id))
    sys.exit(2)
  else:
    #print("dbg553: router_bit_arc_uw:", router_bit_arc_uw)
    if(abs(router_bit_arc_uw)>math.pi):
      print("ERR887: Error in {:s}, the sub smooth corner is englobed!".format(ai_error_msg_id))
      sys.exit(2)
      enlarge_status = 2
    elif(abs(router_bit_arc_uw)<radian_epsilon):
      print("WARN877: Error in {:s}, the corner is too flat to be enlarged with this router_bit radius!".format(ai_error_msg_id))
      enlarge_status = 2
      enlarge_type_request_1 = 1
      enlarge_type_request_2 = 1
    else:
      corner_orientation = math.copysign(1, router_bit_arc_uw)
      enlarge_status = 1
      (SClx, SCly, SCkS, lSC, xSC) = line_equation((SX,SY), (CX,CY), ai_error_msg_id)
      (FX,FY, line_circle_intersection_status) = line_circle_intersection((SClx, SCly, SCkS), (CX,CY),ai_router_bit_request, (SX,SY), xSC+math.pi, ai_error_msg_id)
      if(line_circle_intersection_status==2):
        print("ERR639: Error in {:s} with the line_circle_intersection_status!".format(ai_error_msg_id))
        sys.exit(2)
      (GX,GY, SCkG) = line_distance_point((FX,FY), (CX,CY), -1*corner_orientation*ai_router_bit_request, ai_error_msg_id)
      (HX,HY, SCkH) = line_distance_point((FX,FY), (CX,CY),  1*corner_orientation*ai_router_bit_request, ai_error_msg_id)
      (IX,IY, R1, uw1, u1, w1) = arc_center_radius_angles(ai_pre_point, ai_pre_middle, ai_current_point, ai_error_msg_id)
      (JX,JY, R2, uw2, u2, w2) = arc_center_radius_angles(ai_current_point, ai_post_middle, ai_post_point, ai_error_msg_id)
      (MX,MY, line_circle_intersection_status) = line_circle_intersection((SClx, SCly, SCkG), (IX,IY),R1, ((GX+SX)/2,(GY+SY)/2), xSC+math.pi, ai_error_msg_id)
      #print("dbg321: MX={:0.2f}  MY={:0.2f}  line_circle_intersection={:0.2f}".format(MX, MY, line_circle_intersection_status))
      if(line_circle_intersection_status==2):
        print("ERR636: Error in {:s} with the line_circle_intersection_status!".format(ai_error_msg_id))
        sys.exit(2)
      (NX,NY, line_circle_intersection_status) = line_circle_intersection((SClx, SCly, SCkH), (JX,JY),R2, ((HX+SX)/2,(HY+SY)/2), xSC+math.pi, ai_error_msg_id)
      #print("dbg322: NX={:0.2f}  NY={:0.2f}  line_circle_intersection={:0.2f}".format(NX, NY, line_circle_intersection_status))
      if(line_circle_intersection_status==2):
        print("ERR633: Error in {:s} with the line_circle_intersection_status!".format(ai_error_msg_id))
        #print("dbg513: G={:0.2f} {:0.2f}  H={:0.2f} {:0.2f}".format(GX, GY, HX, HY))
        #sys.exit(2)
        enlarge_status = 2
      # check if arc-arc intersection must be calculated
      tmp_MG_deep_x = (MX-GX)*(CX-FX)
      tmp_MG_deep_y = (MY-GY)*(CY-FY)
      #print("dbg301: tmp_MG_deep_x:", tmp_MG_deep_x)
      #print("dbg302: tmp_MG_deep_y:", tmp_MG_deep_y)
      MG_deep = math.copysign(1, tmp_MG_deep_x)
      if(abs(tmp_MG_deep_x)<abs(tmp_MG_deep_y)):
        MG_deep = math.copysign(1, tmp_MG_deep_y)
        #print("dbg303: use Y to get MG_deep")
      tmp_NH_deep_x = (NX-HX)*(CX-FX)
      tmp_NH_deep_y = (NY-HY)*(CY-FY)
      #print("dbg311: tmp_NH_deep_x:", tmp_NH_deep_x)
      #print("dbg312: tmp_NH_deep_y:", tmp_NH_deep_y)
      NH_deep = math.copysign(1, tmp_NH_deep_x)
      if(abs(tmp_NH_deep_x)<abs(tmp_NH_deep_y)):
        NH_deep = math.copysign(1, tmp_NH_deep_y)
        #print("dbg313: use Y to get NH_deep")
      MKX = MX
      MKY = MY
      NLX = NX
      NLY = NY
      enlarge_type_request_1 = 3
      enlarge_type_request_2 = 3
      # compute the arc-arc intersection
      if(MG_deep>0):
        (KX,KY, triangulation_status) = triangulation((IX,IY),R1,(FX,FY),ai_router_bit_request,(GX,GY), xSC+math.pi, ai_error_msg_id)
        if(triangulation_status==2):
          print("ERR536: Error in {:s} with the triangulation_status!".format(ai_error_msg_id))
          sys.exit(2)
        MKX = KX
        MKY = KY
        enlarge_type_request_1 = 2
      if(NH_deep>0):
        (LX,LY, triangulation_status) = triangulation((JX,JY),R2,(FX,FY),ai_router_bit_request,(HX,HY), xSC+math.pi, ai_error_msg_id)
        if(triangulation_status==2):
          print("ERR533: Error in {:s} with the triangulation_status!".format(ai_error_msg_id))
          sys.exit(2)
        NLX = LX
        NLY = LY
        enlarge_type_request_2 = 2
      lMN=math.sqrt((NLX-MKX)**2+(NLY-MKY)**2)
      if(lMN<radian_epsilon):
        print("WARN667: Warning in {:s}, the angle is too flat to be enlarged!".format(ai_error_msg_id))
        enlarge_type_request_1 = 1
        enlarge_type_request_2 = 1
  # return
  r_sub_enlarge_corner_arc_arc = ((MKX,MKY), (GX,GY), (CX,CY), (HX,HY), (NLX,NLY), enlarge_type_request_1, enlarge_type_request_2, enlarge_status)
  #print("dbg588: r_sub_enlarge_corner_arc_arc:", r_sub_enlarge_corner_arc_arc)
  return(r_sub_enlarge_corner_arc_arc)

def curve_arc(ai_AX, ai_AY, ai_CX, ai_CY, ai_At, ai_router_bit_request, ai_error_msg_id):
  """ Given the start-point (A), the end-point (C) and the arc tangent (At) at (A), return the middle-point (B) and the arc tangent (Ct) at (C)
  """
  # use to check if the calculation is possible
  radian_epsilon = math.pi/1000
  # line equation of (AC)
  (AClx, ACly, ACk, lAC, xAC) = line_equation((ai_AX, ai_AY), (ai_CX, ai_CY), ai_error_msg_id)
  # calcultion of I the middle of [AC]
  IX = (ai_AX+ai_CX)/2
  IY = (ai_AY+ai_CY)/2
  # line equation of the bisection (OI) of [AC]
  OIlx = ACly
  OIly = -1*AClx
  OIk = -1*(OIlx*IX+OIly*IY)
  # line equation of (OA)
  # Tangent(A) equation (sin(At), -cos(At), K)
  # OA is perpendicular to Tangent(A)
  OAlx = math.cos(ai_At)
  OAly = math.sin(ai_At)
  OAk = -1*(OAlx*ai_AX+OAly*ai_AY)
  # O intersection of (OI) and (OA). It's the center of the arc
  (OX, OY, line_line_intersection_status) = line_line_intersection((OIlx, OIly, OIk),(OAlx, OAly, OAk), ai_error_msg_id)
  if(line_line_intersection_status==2):
    print("ERR374: Error in {:s}, the tangent and AC are collinear!".format(ai_error_msg_id))
    sys.exit(2)
  # verification of the distance OA and OC
  lOA = math.sqrt((ai_AX-OX)**2+(ai_AY-OY)**2)
  lOC = math.sqrt((ai_CX-OX)**2+(ai_CY-OY)**2)
  #print("dbg354: ai_AX={:0.2f}  ai_AY={:0.2f}".format(ai_AX, ai_AY))
  #print("dbg355: ai_CX={:0.2f}  ai_CY={:0.2f}".format(ai_CX, ai_CY))
  #print("dbg356: IX={:0.2f}  IY={:0.2f}".format(IX, IY))
  #print("dbg357: OX={:0.2f}  OY={:0.2f}".format(OX, OY))
  if(abs(lOC-lOA)>radian_epsilon):
    print("ERR375: Error in {:s}, O is not equidistant from A and C! lOA={:0.2f} l_OC={:0.2f}".format(ai_error_msg_id, lOA, lOC))
    sys.exit(2)
  if(lOA<ai_router_bit_request):
    print("WARN446: Warning in {:s}, the radius_of_curvature is smaller than the router_bit_request! lOA={:0.2f} rbr={:0.2f}".format(ai_error_msg_id, lOA, ai_router_bit_request))
  # calculation of the angle (Ox, OA)
  xOA = math.atan2((ai_AY-OY)/lOA, (ai_AX-OX)/lOA)
  # calculation of the angle (Ox, OC)
  xOC = math.atan2((ai_CY-OY)/lOC, (ai_CX-OX)/lOC)
  # arc orientation = sign of the angle (At,AC)
  AtAC = math.fmod(xAC-ai_At+5*math.pi, 2*math.pi) - math.pi
  # calculation of the angle AOC
  AOC = 0
  if(AtAC>0):
    AOC = math.fmod(xOC-xOA+4*math.pi, 2*math.pi)
  else:
    AOC = -1*math.fmod(xOA-xOC+4*math.pi, 2*math.pi)
  if(abs(AOC)<radian_epsilon):
    print("WARN776: Warning in {:s}, the angle AOC is really small!".format(ai_error_msg_id))
  # calculation of B
  xOB = xOA + AOC/2
  BX = OX+lOA*math.cos(xOB)
  BY = OY+lOA*math.sin(xOB)
  # calculation of the tangent inclination Ct in C
  Ct = xOC + math.copysign(math.pi/2, AtAC)
  # return
  r_curve_arc = (BX, BY, Ct)
  return(r_curve_arc)

def unstable_smooth_outline_b_curve(ai_polyline, ai_initial_tangent, ai_precision, ai_router_bit_request, ai_error_msg_id):
  """
  This function computes a serie of N arcs that pass through the (N+1) points defined by the N line-segments of ai_polyline.
  ai_polyline is an outline of format B, that contains only lines (no arcs) (list of list of 2 floats).
  ai_initial_tangent is the tangent of the wished curve at the first-point of the outline ai_polyline.
  ai_precision defined when a line should be generated instead of an arc because the tangent and the new segment are almost collinear.
  ai_router_bit_request is just used to warn if the radius_of_curvature is smaller than the router_bit. Because this function can not know on which side of the outline is the material, those warnings might be irrelevant. If you don't want this feature, just set it to 0, this disables these warnings.
  ai_error_msg_id is a string, that can help you to track bugs and erros.
  The function returns an outline of format B containing only arcs
  """
  # define the angle precision to know when to use a line instead of an arc
  #radian_epsilon = 1/1000.0
  radian_epsilon = ai_precision
  # check if the input outline is closed
  outline_closed = False
  if((ai_polyline[0][0]==ai_polyline[-1][-2])and(ai_polyline[0][1]==ai_polyline[-1][-1])):
  #if((abs(ai_polyline[0][0]-ai_polyline[-1][-2])<radian_epsilon)and(abs(ai_polyline[0][1]==ai_polyline[-1][-1])<radian_epsilon)):
    outline_closed = True
    print("WARN994: Warning in {:s}, the curve to smooth is closed and this will be ignored by the unstable_smooth_curve() function".format(ai_error_msg_id))
  # number of corners and segments
  point_nb = len(ai_polyline)
  # check the outline point number
  if(point_nb<2):
    print("ERR209: Error in {:s}, the number of points must be bigger than 2. Currently: {:d}".format(ai_error_msg_id, point_nb))
    sys.exit(2)
  # check if the first point is valid
  first_point_len = len(ai_polyline[0])
  if(first_point_len!=2):
    print("ERR219: Error in {:s}, the first-point of ai_polyline must have exactly 2 elements. Currently: {:d}".format(ai_error_msg_id, first_point_len))
    sys.exit(2)
  # processing initialization
  ti = ai_initial_tangent
  r_outline = []
  r_outline.append(ai_polyline[0]) # first-point
  # processing incrementation
  for i in range(point_nb-1):
    # error message
    i_error_msg_id = "{:s}.{:d}".format(ai_error_msg_id, i)
    # check the validity of the new segment
    segment_len = len(ai_polyline[i+1])
    if(segment_len!=2):
      print("ERR229: Error in {:s}, the ai_polyline segment length must be exactly 2. Currently: {:d}".format(i_error_msg_id, segment_len))
      sys.exit(2)
    # geometrical data
    AX = ai_polyline[i][0]
    AY = ai_polyline[i][1]
    CX = ai_polyline[i+1][0]
    CY = ai_polyline[i+1][1]
    # calculation of the inclination of AC
    lAC = math.sqrt((CX-AX)**2+(CY-AY)**2)
    iAC = math.atan2((CY-AY)/lAC, (CX-AX)/lAC)
    rti = math.fmod(ti-iAC+5*math.pi, 2*math.pi)-math.pi # angle (AC, tangent) between [-pi,pi]
    if(abs(rti)>math.pi-radian_epsilon):
      print("ERR239: Error in {:s}, AC and the tangent Ti are collinear and in opposite direction. iAC={:0.2f}  ti={:0.2f}".format(i_error_msg_id, iAC, ti))
      sys.exit(2)
    if(abs(rti)>math.pi/3):
      print("WARN249: Warning in {:s}, AC and the tangent Ti are doing a large angle. Add itermediate points to remove this warning. iAC={:0.2f}  ti={:0.2f}".format(i_error_msg_id, iAC, ti))
    if(abs(rti)<radian_epsilon):
      print("WARN259: Warning in {:s}, AC and the tangent Ti are almost identical. A line is generated for this segment. iAC={:0.2f}  ti={:0.2f}".format(i_error_msg_id, iAC, ti))
      r_outline.append([CX, CY]) # create a line-segment
      ti = iAC
    else:
      (BX, BY, nti) = curve_arc(AX, AY, CX, CY, ti, ai_router_bit_request, i_error_msg_id)
      r_outline.append([BX, BY, CX, CY]) # create an arc-segment
      ti = nti
  print("dbg536: the last tangent inclination ti: {:0.2f}".format(ti))
  # return
  return(r_outline)

def sub_smooth_outline_c_curve(ai_polyline, ai_precision, ai_router_bit_request, ai_error_msg_id):
  """
  This function computes a serie of 2N arcs that pass through the (N+1) points defined by the N line-segments of ai_polyline.
  ai_polyline is an outline of format C, (list of list of 3 floats).
  ai_precision defined when a line should be generated instead of an arc because the tangent and the new segment are almost collinear.
  ai_router_bit_request is just used to warn if the radius_of_curvature is smaller than the router_bit. Because this function can not know on which side of the outline is the material, those warnings might be irrelevant. If you don't want this feature, just set it to 0, this disables these warnings.
  ai_error_msg_id is a string, that can help you to track bugs and erros.
  The function returns an outline of format B containing only arcs
  """
  # define the angle precision to know when to use a line instead of an arc
  #radian_epsilon = 1/1000.0
  radian_epsilon = ai_precision
  # check if the input outline is closed
  outline_closed = False
  if((ai_polyline[0][0]==ai_polyline[-1][0])and(ai_polyline[0][1]==ai_polyline[-1][1])):
  #if((abs(ai_polyline[0][0]-ai_polyline[-1][-2])<radian_epsilon)and(abs(ai_polyline[0][1]==ai_polyline[-1][-1])<radian_epsilon)):
    outline_closed = True
    #print("WARN694: Warning in {:s}, the curve to smooth is closed and this will be ignored by the unstable_smooth_curve() function".format(ai_error_msg_id))
  # number of corners and segments
  point_nb = len(ai_polyline)
  # check the outline point number
  if(point_nb<2):
    print("ERR609: Error in {:s}, the number of points must be bigger than 2. Currently: {:d}".format(ai_error_msg_id, point_nb))
    sys.exit(2)
  # check if the first point is valid
  first_point_len = len(ai_polyline[0])
  if(first_point_len!=3):
    print("ERR619: Error in {:s}, the first-point of ai_polyline must have exactly 3 elements. Currently: {:d}".format(ai_error_msg_id, first_point_len))
    sys.exit(2)
  # processing initialization
  r_outline = []
  r_outline.append((ai_polyline[0][0], ai_polyline[0][1])) # first-point
  # processing incrementation
  for i in range(point_nb-1):
    # error message
    i_error_msg_id = "{:s}.{:d}".format(ai_error_msg_id, i)
    # check the validity of the new segment
    segment_len = len(ai_polyline[i+1])
    if(segment_len!=3):
      print("ERR629: Error in {:s}, the ai_polyline segment length must be exactly 3. Currently: {:d}".format(i_error_msg_id, segment_len))
      sys.exit(2)
    # geometrical data
    AX = ai_polyline[i][0]
    AY = ai_polyline[i][1]
    xAt = ai_polyline[i][2]
    EX = ai_polyline[i+1][0]
    EY = ai_polyline[i+1][1]
    xEt = ai_polyline[i+1][2]
    # calculation of the inclination of AE
    lAE = math.sqrt((EX-AX)**2+(EY-AY)**2)
    xAE = math.atan2((EY-AY)/lAE, (EX-AX)/lAE)
    # calculation of the inclination of AC
    AtAE = math.fmod(xAt-xAE+5*math.pi, 2*math.pi)-math.pi # angle (AE, tangent) between [-pi,pi]
    xAC = math.fmod(xAE + AtAE/2 + 5*math.pi, 2*math.pi) - math.pi
    AClx = math.sin(xAC)
    ACly = -1*math.cos(xAC)
    ACk = -1*(AClx*AX+ACly*AY)
    # calculation of the inclination of EC
    EtEA = math.fmod(xEt-xAE+5*math.pi, 2*math.pi)-math.pi # angle (AE, tangent) between [-pi,pi]
    xEC = math.fmod(xAE+math.pi + EtEA/2 + 5*math.pi, 2*math.pi) - math.pi
    EClx = math.sin(xEC)
    ECly = -1*math.cos(xEC)
    ECk = -1*(EClx*EX+ECly*EY)
    # check if the segment must be an arc or a line
    if(abs(AtAE)>math.pi/2):
      print("ERR639: Error in {:s}, the angle between AC and the tangent xAt is larger than pi/2. It doesn't look like a feasible curbe. xAC={:0.2f}  xAt={:0.2f}".format(i_error_msg_id, xAC, xAt))
      sys.exit(2)
    if(abs(EtEA)>math.pi/2):
      print("ERR638: Error in {:s}, the angle between EC and the tangent et is larger than pi/2. It doesn't look like a feasible curbe. xEC={:0.2f}  xEt={:0.2f}".format(i_error_msg_id, xEC, xEt))
      sys.exit(2)
    if(abs(AtAE)>math.pi/3):
      print("WARN649: Warning in {:s}, AC and the tangent xAt are doing a large angle. Add itermediate points to remove this warning. xAC={:0.2f}  xAt={:0.2f}".format(i_error_msg_id, xAC, xAt))
    if(abs(EtEA)>math.pi/3):
      print("WARN648: Warning in {:s}, EC and the tangent xEt are doing a large angle. Add itermediate points to remove this warning. xEC={:0.2f}  xEt={:0.2f}".format(i_error_msg_id, xEC, xEt))
    if((abs(AtAE)<radian_epsilon)or(abs(EtEA)<radian_epsilon)):
      print("WARN659: Warning in {:s}, (xAC, xAt) or (xEC, xEt) are almost identical. A line is generated for this segment. xAC={:0.2f}  xAt={:0.2f} xEC={:0.2f}  xEt={:0.2f} ".format(i_error_msg_id, xAC, xAt, xEC, xEt))
      r_outline.append((EX, EY)) # create a line-segment
    elif((AtAE*EtEA)>0):
      print("WARN669: Warning in {:s}, xAt and xEt are not one the side of (AE). It look like an inflexion. A line is generated for this segment. xAC={:0.2f}  xAt={:0.2f} xEC={:0.2f} xEt={:0.2f} ".format(i_error_msg_id, xAC, xAt, xEC, xEt))
      r_outline.append((EX, EY)) # create a line-segment
    else:
      # C intersection of (AC) and (EC). it is the junction point between the two arcs
      (CX, CY, line_line_intersection_status) = line_line_intersection((AClx, ACly, ACk),(EClx, ECly, ECk), ai_error_msg_id)
      if(line_line_intersection_status==2):
        print("ERR324: Error in {:s}, AC and EC are collinear!".format(ai_error_msg_id))
        sys.exit(2)
      (BX, BY, xCt) = curve_arc(AX, AY, CX, CY, xAt, ai_router_bit_request, i_error_msg_id)
      if(abs(math.fmod(xCt-xAE+5*math.pi, 2*math.pi)-math.pi)>radian_epsilon):
        print("ERR325: Error in {:s}, the first arc tangent in C is not parallel to AE! xCt={:0.2f} xAE={:0.2f}".format(ai_error_msg_id, xCt, xAE))
        sys.exit(2)
      (DX, DY, xEt2) = curve_arc(CX, CY, EX, EY, xAE, ai_router_bit_request, i_error_msg_id)
      if(abs(math.fmod(xEt2-xEt+5*math.pi, 2*math.pi)-math.pi)>radian_epsilon):
        print("ERR326: Error in {:s}, the second arc tangent in E is different from xEt! xEt={:0.2f} xEt2={:0.2f}".format(ai_error_msg_id, xEt, xEt2))
        sys.exit(2)
      r_outline.append((BX, BY, CX, CY)) # create the first arc-segment
      r_outline.append((DX, DY, EX, EY)) # create the second arc-segment
  # return
  return(r_outline)


